from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ldap3


class CERNLDAPSyncPlugin(object):
    """Synchronization plugin for mapping new users to CERN accounts.

    This plugin results in new users having two additional fields in the CS,
    CERNAccountType and PrimaryCERNAccount. If the new nickname does not have a
    corresponding CERN account it will be rejected.
    """

    def __init__(self):
        """Initialise the plugin and underlying LDAP connection."""
        self._server = ldap3.Server("ldap://xldap.cern.ch")
        self._connection = ldap3.Connection(self._server, client_strategy=ldap3.SAFE_SYNC, auto_bind=True)

    def verifyAndUpdateUserInfo(self, username, userDict):
        """Add the "CERNAccountType" and "PrimaryCERNAccount" values to the CS attributes.

        :param username: DIRAC name of the user to be added
        :param userDict: user information collected by the VOMS2CSAgent
        :returns: None
        :raise ValueError: if no corresponding CERN account is found.
        """
        attributes = self._getUserInfo(username)
        cernAccountType = attributes["cernAccountType"]
        userDict["CERNAccountType"] = cernAccountType[0]
        if cernAccountType == ["Primary"]:
            userDict["PrimaryCERNAccount"] = username
        else:
            userDict["PrimaryCERNAccount"] = self._findOwnerAccountName(username, attributes)

    def _findOwnerAccountName(self, username, attributes):
        """Find the owner account from a CERN LDAP entry.

        :param username: DIRAC name of the user to be added
        :param attributes: output of ``_getUserInfo`` for ``username``
        :returns: The name of the owning CERN account
        """
        owners = attributes["cernAccountOwner"]
        if len(owners) != 1:
            raise ValueError("Expected exactly one cernAccountOwner for %s but found %s" % (username, len(owners)))
        commonNames = [v for k, v, _ in ldap3.utils.dn.parse_dn(owners[0]) if k == "CN"]
        if len(commonNames) != 1:
            raise ValueError(
                "Expected exactly one common name in the cernAccountOwner of %s but found %s"
                % (username, len(commonNames))
            )
        primaryAccountName = commonNames[0]
        primaryAttributes = self._getUserInfo(primaryAccountName)
        if primaryAttributes["cernAccountType"] != ["Primary"]:
            raise ValueError("Something is very wrong!")
        return primaryAccountName

    def _getUserInfo(self, commonName):
        """Query the CERN LDAP server for the given ``commonName``.

        :param commonName: Common Name of an account known to CERN
        :returns: ``dict`` of the account attributes from LDAP
        :raise ValueError: if no corresponding CERN account is found
        """
        status, result, response, _ = self._connection.search(
            "OU=Users,OU=Organic Units,DC=cern,DC=ch",
            "(CN=%s)" % commonName,
            attributes=["cernAccountOwner", "cernAccountType"],
        )
        if not status:
            raise ValueError("Bad status from LDAP search: %s" % result)
        if len(response) != 1:
            raise ValueError("Expected exactly one match for CN=%s but found %s" % (commonName, len(response)))
        # https://github.com/PyCQA/pylint/issues/4148
        return response[0]["attributes"]  # pylint: disable=unsubscriptable-object



""" Used by the executors for dispatching events (IIUC)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading
import time

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.ReturnValues import isReturnStructure
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.FrameworkSystem.Client.MonitoringClient import MonitoringClient
from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter


class ExecutorState(object):
    def __init__(self, log=False):
        if log:
            self.__log = log
        else:
            self.__log = gLogger
        self.__lock = threading.Lock()
        self.__typeToId = {}
        self.__maxTasks = {}
        self.__execTasks = {}
        self.__taskInExec = {}

    def _internals(self):
        return {
            "type2id": dict(self.__typeToId),
            "maxTasks": dict(self.__maxTasks),
            "execTasks": dict(self.__execTasks),
            "tasksInExec": dict(self.__taskInExec),
            "locked": self.__lock.locked(),  # pylint: disable=no-member
        }

    def addExecutor(self, eId, eTypes, maxTasks=1):
        self.__lock.acquire()
        try:
            self.__maxTasks[eId] = max(1, maxTasks)
            if eId not in self.__execTasks:
                self.__execTasks[eId] = set()
            if not isinstance(eTypes, (list, tuple)):
                eTypes = [eTypes]
            for eType in eTypes:
                if eType not in self.__typeToId:
                    self.__typeToId[eType] = set()
                self.__typeToId[eType].add(eId)
        finally:
            self.__lock.release()

    def removeExecutor(self, eId):
        self.__lock.acquire()
        try:
            tasks = []
            for eType in self.__typeToId:
                if eId in self.__typeToId[eType]:
                    self.__typeToId[eType].remove(eId)
            for taskId in self.__execTasks[eId]:
                self.__taskInExec.pop(taskId)
                tasks.append(taskId)
            self.__execTasks.pop(eId)
            self.__maxTasks.pop(eId)
            return tasks
        finally:
            self.__lock.release()

    def getTasksForExecutor(self, eId):
        try:
            return set(self.__execTasks[eId])
        except KeyError:
            return set()

    def full(self, eId):
        try:
            return len(self.__execTasks[eId]) >= self.__maxTasks[eId]
        except KeyError:
            return True

    def freeSlots(self, eId):
        try:
            return self.__maxTasks[eId] - len(self.__execTasks[eId])
        except KeyError:
            return 0

    def getFreeExecutors(self, eType):
        execs = {}
        try:
            eids = self.__typeToId[eType]
        except KeyError:
            return execs
        try:
            for eid in eids:
                freeSlots = self.freeSlots(eid)
                if freeSlots:
                    execs[eid] = freeSlots
        except RuntimeError:
            pass
        return execs

    def getIdleExecutor(self, eType):
        idleId = None
        maxFreeSlots = 0
        try:
            # Work on a copy of self.__typeToId[eType] to race conditions causing it's size to change while iterating
            for eId in list(self.__typeToId[eType]):
                freeSlots = self.freeSlots(eId)
                if freeSlots > maxFreeSlots:
                    maxFreeSlots = freeSlots
                    idleId = eId
        except KeyError:
            pass
        return idleId

    def addTask(self, eId, taskId):
        self.__lock.acquire()
        try:
            try:
                self.__taskInExec[taskId] = eId
                self.__execTasks[eId].add(taskId)
                return len(self.__execTasks[eId])
            except KeyError:
                return 0
        finally:
            self.__lock.release()

    def getExecutorOfTask(self, taskId):
        try:
            return self.__taskInExec[taskId]
        except KeyError:
            return None

    def removeTask(self, taskId, eId=None):
        self.__lock.acquire()
        try:
            try:
                if eId is None:
                    eId = self.__taskInExec[taskId]
                self.__execTasks[eId].remove(taskId)
                self.__taskInExec.pop(taskId)
                return True
            except KeyError:
                return False
        finally:
            self.__lock.release()


class ExecutorQueues(object):
    def __init__(self, log=False):
        if log:
            self.__log = log
        else:
            self.__log = gLogger
        self.__lock = threading.Lock()
        self.__queues = {}
        self.__lastUse = {}
        self.__taskInQueue = {}

    def _internals(self):
        return {
            "queues": dict(self.__queues),
            "lastUse": dict(self.__lastUse),
            "taskInQueue": dict(self.__taskInQueue),
            "locked": self.__lock.locked(),  # pylint: disable=no-member
        }

    def getExecutorList(self):
        return [eType for eType in self.__queues]

    def pushTask(self, eType, taskId, ahead=False):
        self.__log.verbose("Pushing task %s into waiting queue for executor %s" % (taskId, eType))
        self.__lock.acquire()
        try:
            if taskId in self.__taskInQueue:
                if self.__taskInQueue[taskId] != eType:
                    errMsg = "Task %s cannot be queued because it's already queued for %s" % (
                        taskId,
                        self.__taskInQueue[taskId],
                    )
                    self.__log.fatal(errMsg)
                    return 0
                return len(self.__queues[eType])
            if eType not in self.__queues:
                self.__queues[eType] = []
            self.__lastUse[eType] = time.time()
            if ahead:
                self.__queues[eType].insert(0, taskId)
            else:
                self.__queues[eType].append(taskId)
            self.__taskInQueue[taskId] = eType
            return len(self.__queues[eType])
        finally:
            self.__lock.release()

    def popTask(self, eTypes):
        if not isinstance(eTypes, (list, tuple)):
            eTypes = [eTypes]
        self.__lock.acquire()
        for eType in eTypes:
            try:
                taskId = self.__queues[eType].pop(0)
                del self.__taskInQueue[taskId]
                # Found! release and return!
                self.__lock.release()
                self.__lastUse[eType] = time.time()
                self.__log.verbose("Popped task %s from executor %s waiting queue" % (taskId, eType))
                return (taskId, eType)
            except IndexError:
                continue
            except KeyError:
                continue
        self.__lock.release()
        # Not found. release and return None
        return None

    def getState(self):
        self.__lock.acquire()
        try:
            qInfo = {}
            for qName in self.__queues:
                qInfo[qName] = list(self.__queues[qName])
        finally:
            self.__lock.release()
        return qInfo

    def deleteTask(self, taskId):
        self.__log.verbose("Deleting task %s from waiting queues" % taskId)
        self.__lock.acquire()
        try:
            try:
                eType = self.__taskInQueue[taskId]
                del self.__taskInQueue[taskId]
                self.__lastUse[eType] = time.time()
            except KeyError:
                return False
            try:
                iPos = self.__queues[eType].index(taskId)
            except ValueError:
                return False
            del self.__queues[eType][iPos]
            return True
        finally:
            self.__lock.release()

    def waitingTasks(self, eType):
        self.__lock.acquire()
        try:
            try:
                return len(self.__queues[eType])
            except KeyError:
                return 0
        finally:
            self.__lock.release()


class ExecutorDispatcherCallbacks(object):
    def cbDispatch(self, taskId, taskObj, pathExecuted):
        return S_ERROR("No dispatch callback defined")

    def cbSendTask(self, taskId, taskObj, eId, eType):
        return S_ERROR("No send task callback defined")

    def cbDisconectExecutor(self, eId):
        return S_ERROR("No disconnect callback defined")

    def cbTaskError(self, taskId, taskObj, errorMsg):
        return S_ERROR("No error callback defined")

    def cbTaskProcessed(self, taskId, taskObj, eType):
        return S_OK()

    def cbTaskFreeze(self, taskId, taskObj, eType):
        return S_OK()


class ExecutorDispatcher(object):
    class ETask(object):
        def __init__(self, taskId, taskObj):
            self.taskId = taskId
            self.taskObj = taskObj
            self.pathExecuted = []
            self.freezeTime = 60
            self.frozenTime = 0
            self.frozenSince = 0
            self.frozenCount = 0
            self.frozenMsg = False
            self.eType = False
            self.sendTime = 0
            self.retries = 0

        def __repr__(self):
            rS = "<ETask %s" % self.taskId
            if self.eType:
                rS += " eType=%s>" % self.eType
            else:
                rS += ">"
            return rS

    def __init__(self, monitor=None):
        """
        :param monitor: good question.... what's meant to be used for monitoring.
            Either a :py:class`DIRAC.FrameworkSystem.Client.MonitoringClient.MonitoringClient` or a
            :py:class`DIRAC.MonitoringSystem.Client.MonitoringReporter.MonitoringReporter`
        """
        self.__idMap = {}
        self.__execTypes = {}
        self.__executorsLock = threading.Lock()
        self.__tasksLock = threading.Lock()
        self.__freezerLock = threading.Lock()
        self.__tasks = {}
        self.__log = gLogger.getSubLogger("ExecMind")
        self.__taskFreezer = []
        self.__queues = ExecutorQueues(self.__log)
        self.__states = ExecutorState(self.__log)
        self.__cbHolder = ExecutorDispatcherCallbacks()
        self.__monitor = None
        if isinstance(monitor, MonitoringClient):
            self.__monitor = monitor
        elif isinstance(monitor, MonitoringReporter):
            self.__monitoringReporter = monitor
        gThreadScheduler.addPeriodicTask(60, self.__doPeriodicStuff)
        # If a task is frozen too many times, send error or forget task?
        self.__failedOnTooFrozen = True
        # If a task fails to properly dispatch, freeze or forget task?
        self.__freezeOnFailedDispatch = True
        # If a task needs to go to an executor that has not connected. Freeze or forget the task?
        self.__freezeOnUnknownExecutor = True
        if self.__monitor:
            self.__monitor.registerActivity(
                "executors", "Executor reactors connected", "Executors", "executors", self.__monitor.OP_MEAN, 300
            )
            self.__monitor.registerActivity(
                "tasks", "Tasks processed", "Executors", "tasks", self.__monitor.OP_RATE, 300
            )
            self.__monitor.registerActivity(
                "taskTime", "Task processing time", "Executors", "seconds", self.__monitor.OP_MEAN, 300
            )

    def setFailedOnTooFrozen(self, value):
        self.__failedOnTooFrozen = value

    def setFreezeOnFailedDispatch(self, value):
        self.__freezeOnFailedDispatch = value

    def setFreezeOnUnknownExecutor(self, value):
        self.__freezeOnUnknownExecutor = value

    def _internals(self):
        return {
            "idMap": dict(self.__idMap),
            "execTypes": dict(self.__execTypes),
            "tasks": sorted(self.__tasks),
            "freezer": list(self.__taskFreezer),
            "queues": self.__queues._internals(),
            "states": self.__states._internals(),
            "locked": {
                "exec": self.__executorsLock.locked(),  # pylint: disable=no-member
                "tasks": self.__tasksLock.locked(),  # pylint: disable=no-member
                "freezer": self.__freezerLock.locked(),  # pylint: disable=no-member
            },
        }

    def setCallbacks(self, callbacksObj):
        if not isinstance(callbacksObj, ExecutorDispatcherCallbacks):
            return S_ERROR("Callbacks object does not inherit from ExecutorDispatcherCallbacks")
        self.__cbHolder = callbacksObj
        return S_OK()

    def __doPeriodicStuff(self):
        self.__unfreezeTasks()
        for eType in self.__execTypes:
            self.__fillExecutors(eType)
        if not self.__monitor:
            return
        eTypes = self.__execTypes
        for eType in eTypes:
            try:
                self.__monitor.addMark("executors-%s" % eType, self.__execTypes[eType])
            except KeyError:
                pass
        self.__monitor.addMark("executors", len(self.__idMap))

    def addExecutor(self, eId, eTypes, maxTasks=1):
        self.__log.verbose("Adding new executor to the pool", "%s: %s" % (eId, ", ".join(eTypes)))
        self.__executorsLock.acquire()
        try:
            if eId in self.__idMap:
                return
            if not isinstance(eTypes, (list, tuple)):
                eTypes = [eTypes]
            self.__idMap[eId] = list(eTypes)
            self.__states.addExecutor(eId, eTypes, maxTasks)
            for eType in eTypes:
                if eType not in self.__execTypes:
                    self.__execTypes[eType] = 0
                    if self.__monitor:
                        self.__monitor.registerActivity(
                            "executors-%s" % eType,
                            "%s executor modules connected" % eType,
                            "Executors",
                            "executors",
                            self.__monitor.OP_MEAN,
                            300,
                        )
                        self.__monitor.registerActivity(
                            "tasks-%s" % eType,
                            "Tasks processed by %s" % eType,
                            "Executors",
                            "tasks",
                            self.__monitor.OP_RATE,
                            300,
                        )
                        self.__monitor.registerActivity(
                            "taskTime-%s" % eType,
                            "Task processing time for %s" % eType,
                            "Executors",
                            "seconds",
                            self.__monitor.OP_MEAN,
                            300,
                        )
                self.__execTypes[eType] += 1
        finally:
            self.__executorsLock.release()
        for eType in eTypes:
            self.__fillExecutors(eType)

    def removeExecutor(self, eId):
        self.__log.info("Removing executor", eId)
        self.__executorsLock.acquire()
        try:
            if eId not in self.__idMap:
                return
            eTypes = self.__idMap.pop(eId)
            for eType in eTypes:
                self.__execTypes[eType] -= 1
            tasksInExec = self.__states.removeExecutor(eId)
            for taskId in tasksInExec:
                try:
                    eTask = self.__tasks[taskId]
                except KeyError:
                    # Task already removed
                    eTask = None
                if eTask and eTask.eType:
                    self.__queues.pushTask(eTask.eType, taskId, ahead=True)
                else:
                    self.__dispatchTask(taskId)
        finally:
            self.__executorsLock.release()
        try:
            self.__cbHolder.cbDisconectExecutor(eId)
        except Exception:
            self.__log.exception("Exception while disconnecting agent %s" % eId)
        for eType in eTypes:
            self.__fillExecutors(eType)

    def __freezeTask(self, taskId, errMsg, eType=False, freezeTime=60):
        self.__log.verbose("Freezing task", taskId)
        self.__freezerLock.acquire()
        try:
            if taskId in self.__taskFreezer:
                return False
            try:
                eTask = self.__tasks[taskId]
            except KeyError:
                return False
            eTask.freezeTime = freezeTime
            eTask.frozenMessage = errMsg
            eTask.frozenSince = time.time()
            eTask.frozenCount += 1
            eTask.eType = eType
            isFrozen = False
            if eTask.frozenCount < 10:
                self.__taskFreezer.append(taskId)
                isFrozen = True
        finally:
            self.__freezerLock.release()
        if not isFrozen:
            self.removeTask(taskId)
            if self.__failedOnTooFrozen:
                self.__cbHolder.cbTaskError(
                    taskId, eTask.taskObj, "Retried more than 10 times. Last error: %s" % errMsg
                )
            return False
        return True

    def __isFrozen(self, taskId):
        return taskId in self.__taskFreezer

    def __removeFromFreezer(self, taskId):
        self.__freezerLock.acquire()
        try:
            try:
                iP = self.__taskFreezer.index(taskId)
            except ValueError:
                return False
            self.__taskFreezer.pop(iP)
            try:
                eTask = self.__tasks[taskId]
            except KeyError:
                return False
            eTask.frozenTime += time.time() - eTask.frozenSince
        finally:
            self.__freezerLock.release()
        return True

    def __unfreezeTasks(self, eType=False):
        iP = 0
        while iP < len(self.__taskFreezer):
            self.__freezerLock.acquire()
            try:
                try:
                    taskId = self.__taskFreezer[iP]
                except IndexError:
                    return
                try:
                    eTask = self.__tasks[taskId]
                except KeyError:
                    self.__log.notice("Removing task %s from the freezer. Somebody has removed the task" % taskId)
                    self.__taskFreezer.pop(iP)
                    continue
                # Current taskId/eTask is the one to defrost
                if eType and eType != eTask.eType:
                    iP += 1
                    continue
                if time.time() - eTask.frozenSince < eTask.freezeTime:
                    iP += 1
                    continue
                self.__taskFreezer.pop(iP)
            finally:
                self.__freezerLock.release()
            # Out of the lock zone to minimize zone of exclusion
            eTask.frozenTime += time.time() - eTask.frozenSince
            self.__log.verbose("Unfreezed task %s" % taskId)
            self.__dispatchTask(taskId, defrozeIfNeeded=False)

    def __addTaskIfNew(self, taskId, taskObj):
        self.__tasksLock.acquire()
        try:
            if taskId in self.__tasks:
                self.__log.verbose("Task %s was already known" % taskId)
                return False
            self.__tasks[taskId] = ExecutorDispatcher.ETask(taskId, taskObj)
            self.__log.verbose("Added task %s" % taskId)
            return True
        finally:
            self.__tasksLock.release()

    def getTask(self, taskId):
        try:
            return self.__tasks[taskId].taskObj
        except KeyError:
            return None

    def __dispatchTask(self, taskId, defrozeIfNeeded=True):
        self.__log.verbose("Dispatching task %s" % taskId)
        # If task already in executor skip
        if self.__states.getExecutorOfTask(taskId):
            return S_OK()
        self.__removeFromFreezer(taskId)

        result = self.__getNextExecutor(taskId)

        if not result["OK"]:
            self.__log.warn("Error while calling dispatch callback", result["Message"])
            if self.__freezeOnFailedDispatch:
                if self.__freezeTask(taskId, result["Message"]):
                    return S_OK()
                return result
            taskObj = self.getTask(taskId)
            self.removeTask(taskId)
            self.__cbHolder.cbTaskError(taskId, taskObj, "Could not dispatch task: %s" % result["Message"])
            return S_ERROR("Could not add task. Dispatching task failed")

        eType = result["Value"]
        if not eType:
            self.__log.verbose("No more executors for task %s" % taskId)
            return self.removeTask(taskId)

        self.__log.verbose("Next executor type is %s for task %s" % (eType, taskId))
        if eType not in self.__execTypes:
            if self.__freezeOnUnknownExecutor:
                self.__log.verbose("Executor type %s has not connected. Freezing task %s" % (eType, taskId))
                self.__freezeTask(taskId, "Unknown executor %s type" % eType, eType=eType, freezeTime=0)
                return S_OK()
            self.__log.verbose("Executor type %s has not connected. Forgetting task %s" % (eType, taskId))
            return self.removeTask(taskId)

        self.__queues.pushTask(eType, taskId)
        self.__fillExecutors(eType, defrozeIfNeeded=defrozeIfNeeded)
        return S_OK()

    def __taskProcessedCallback(self, taskId, taskObj, eType):
        try:
            result = self.__cbHolder.cbTaskProcessed(taskId, taskObj, eType)
        except Exception:
            self.__log.exception("Exception while calling taskDone callback")
            return S_ERROR("Exception while calling taskDone callback")

        if not isReturnStructure(result):
            errMsg = "taskDone callback did not return a S_OK/S_ERROR structure"
            self.__log.fatal(errMsg)
            return S_ERROR(errMsg)

        return result

    def __taskFreezeCallback(self, taskId, taskObj, eType):
        try:
            result = self.__cbHolder.cbTaskFreeze(taskId, taskObj, eType)
        except Exception:
            self.__log.exception("Exception while calling taskFreeze callback")
            return S_ERROR("Exception while calling taskFreeze callback")

        if not isReturnStructure(result):
            errMsg = "taskFreeze callback did not return a S_OK/S_ERROR structure"
            self.__log.fatal(errMsg)
            return S_ERROR(errMsg)

        return result

    def __getNextExecutor(self, taskId):
        try:
            eTask = self.__tasks[taskId]
        except KeyError:
            msg = "Task was deleted prematurely while being dispatched"
            self.__log.error(msg, "%s" % taskId)
            return S_ERROR(msg)
        try:
            result = self.__cbHolder.cbDispatch(taskId, eTask.taskObj, tuple(eTask.pathExecuted))
        except Exception:
            self.__log.exception("Exception while calling dispatch callback")
            return S_ERROR("Exception while calling dispatch callback")

        if not isReturnStructure(result):
            errMsg = "Dispatch callback did not return a S_OK/S_ERROR structure"
            self.__log.fatal(errMsg)
            return S_ERROR(errMsg)

        # Assign the next executor type to the task
        if result["OK"]:
            eTask.eType = result["Value"]

        return result

    def getTaskIds(self):
        return list(self.__tasks)

    def getExecutorsConnected(self):
        return dict(self.__execTypes)

    def addTask(self, taskId, taskObj):
        if not self.__addTaskIfNew(taskId, taskObj):
            self.__unfreezeTasks()
            return S_OK()
        return self.__dispatchTask(taskId)

    def removeTask(self, taskId):
        try:
            self.__tasks.pop(taskId)
        except KeyError:
            self.__log.verbose("Task %s is already removed" % taskId)
            return S_OK()
        self.__log.verbose("Removing task %s" % taskId)
        eId = self.__states.getExecutorOfTask(taskId)
        self.__queues.deleteTask(taskId)
        self.__states.removeTask(taskId)
        self.__freezerLock.acquire()
        try:
            try:
                self.__taskFreezer.pop(self.__taskFreezer.index(taskId))
            except KeyError:
                pass
            except ValueError:
                pass
        finally:
            self.__freezerLock.release()
        if eId:
            # Send task to executor if idle
            self.__sendTaskToExecutor(eId, checkIdle=True)
        return S_OK()

    def __taskReceived(self, taskId, eId):
        try:
            eTask = self.__tasks[taskId]
        except KeyError:
            errMsg = "Task %s is not known" % taskId
            self.__log.error("Task is not known", "%s" % taskId)
            return S_ERROR(errMsg)
        if not self.__states.removeTask(taskId, eId):
            self.__log.info("Executor %s says it's processed task %s but it didn't have it" % (eId, taskId))
            return S_OK()
        if eTask.eType not in self.__idMap[eId]:
            errMsg = "Executor type invalid for %s. Redoing task %s" % (eId, taskId)
            self.__log.error("Executor type invalid. Redoing task", "Type %s, Task %s" % (eId, taskId))
            self.removeExecutor(eId)
            self.__dispatchTask(taskId)
            return S_ERROR(errMsg)
        if self.__monitor:
            tTime = time.time() - self.__tasks[taskId].sendTime
            self.__monitor.addMark("taskTime-%s" % eTask.eType, tTime)
            self.__monitor.addMark("taskTime", tTime)
            self.__monitor.addMark("tasks-%s" % eTask.eType, 1)
            self.__monitor.addMark("tasks", 1)
        return S_OK(eTask.eType)

    def freezeTask(self, eId, taskId, freezeTime, taskObj=False):
        result = self.__taskReceived(taskId, eId)
        if not result["OK"]:
            return result
        eType = result["Value"]
        # Executor didn't have the task.
        if not eType:
            # Fill the executor
            self.__sendTaskToExecutor(eId)
            return S_OK()
        if not taskObj:
            taskObj = self.__tasks[taskId].taskObj
        result = self.__taskFreezeCallback(taskId, taskObj, eType)
        if not result["OK"]:
            # Fill the executor
            self.__sendTaskToExecutor(eId)
            return result
        try:
            self.__tasks[taskId].taskObj = taskObj
        except KeyError:
            self.__log.error("Task seems to have been removed while being processed!", "%s" % taskId)
            self.__sendTaskToExecutor(eId, eType)
            return S_OK()
        self.__freezeTask(taskId, "Freeze request by %s executor" % eType, eType=eType, freezeTime=freezeTime)
        self.__sendTaskToExecutor(eId, eType)
        return S_OK()

    def taskProcessed(self, eId, taskId, taskObj=False):
        result = self.__taskReceived(taskId, eId)
        if not result["OK"]:
            return result
        eType = result["Value"]
        # Executor didn't have the task.
        if not eType:
            # Fill the executor
            self.__sendTaskToExecutor(eId)
            return S_OK()
        # Call the done callback
        if not taskObj:
            taskObj = self.__tasks[taskId].taskObj
        result = self.__taskProcessedCallback(taskId, taskObj, eType)
        if not result["OK"]:
            # Fill the executor
            self.__sendTaskToExecutor(eId)
            # Remove the task
            self.removeTask(taskId)
            return result
        # Up until here it's an executor error. From now on it can be a task error
        try:
            self.__tasks[taskId].taskObj = taskObj
            self.__tasks[taskId].pathExecuted.append(eType)
        except KeyError:
            self.__log.error("Task seems to have been removed while being processed!", "%s" % taskId)
            self.__sendTaskToExecutor(eId, eType)
            return S_OK()
        self.__log.verbose("Executor %s processed task %s" % (eId, taskId))
        result = self.__dispatchTask(taskId)
        self.__sendTaskToExecutor(eId, eType)
        return result

    def retryTask(self, eId, taskId):
        if taskId not in self.__tasks:
            errMsg = "Task %s is not known" % taskId
            self.__log.error("Task is not known", "%s" % taskId)
            return S_ERROR(errMsg)
        if not self.__states.removeTask(taskId, eId):
            self.__log.info("Executor %s says it's processed task %s but it didn't have it" % (eId, taskId))
            self.__sendTaskToExecutor(eId)
            return S_OK()
        self.__log.verbose("Executor %s did NOT process task %s, retrying" % (eId, taskId))
        try:
            self.__tasks[taskId].retries += 1
        except KeyError:
            self.__log.error("Task seems to have been removed while waiting for retry!", "%s" % taskId)
            return S_OK()
        return self.__dispatchTask(taskId)

    def __fillExecutors(self, eType, defrozeIfNeeded=True):
        if defrozeIfNeeded:
            self.__log.verbose("Unfreezing tasks for %s" % eType)
            self.__unfreezeTasks(eType)
        self.__log.verbose("Filling %s executors" % eType)
        eId = self.__states.getIdleExecutor(eType)
        while eId:
            result = self.__sendTaskToExecutor(eId, eType)
            if not result["OK"]:
                self.__log.error("Could not send task to executor", "%s" % result["Message"])
            else:
                if not result["Value"]:
                    # No more tasks for eType
                    break
                self.__log.verbose("Task %s was sent to %s" % (result["Value"], eId))
            eId = self.__states.getIdleExecutor(eType)
        self.__log.verbose("No more idle executors for %s" % eType)

    def __sendTaskToExecutor(self, eId, eTypes=False, checkIdle=False):
        if checkIdle and self.__states.freeSlots(eId) == 0:
            return S_OK()
        try:
            searchTypes = list(reversed(self.__idMap[eId]))
        except KeyError:
            self.__log.verbose("Executor %s invalid/disconnected" % eId)
            return S_ERROR("Invalid executor")
        if eTypes:
            if not isinstance(eTypes, (list, tuple)):
                eTypes = [eTypes]
            for eType in reversed(eTypes):
                try:
                    searchTypes.remove(eType)
                except ValueError:
                    pass
                searchTypes.append(eType)
        pData = self.__queues.popTask(searchTypes)
        if pData is None:
            self.__log.verbose("No more tasks for %s" % eTypes)
            return S_OK()
        taskId, eType = pData
        self.__log.verbose("Sending task %s to %s=%s" % (taskId, eType, eId))
        self.__states.addTask(eId, taskId)
        try:
            self.__msgTaskToExecutor(taskId, eId, eType)
        except Exception:
            self.__log.exception("Exception while sending task to executor")
            self.__queues.pushTask(eType, taskId, ahead=False)
            self.__states.removeTask(taskId)
            return S_ERROR("Exception while sending task to executor")
        return S_OK(taskId)

    def __msgTaskToExecutor(self, taskId, eId, eType):
        self.__tasks[taskId].sendTime = time.time()
        result = self.__cbHolder.cbSendTask(taskId, self.__tasks[taskId].taskObj, eId, eType)
        if not isReturnStructure(result):
            errMsg = "Send task callback did not send back an S_OK/S_ERROR structure"
            self.__log.fatal(errMsg)
            raise ValueError(errMsg)
        if not result["OK"]:
            self.__log.error("Failed to cbSendTask", "%r" % result)
            raise RuntimeError(result)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import unittest

from DIRAC.tests.Utilities.utils import cleanTestDir

from DIRAC import gLogger
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.DataManagementSystem.Client.DataManager import DataManager


class IntegrationTest(unittest.TestCase):
    """Base class for the integration and regression tests"""

    def setUp(self):
        cleanTestDir()
        self.dirac = Dirac()
        gLogger.setLevel("DEBUG")

    def tearDown(self):
        #    cleanTestDir()
        pass


class FailingUserJobTestCase(IntegrationTest):
    """Base class for the faing jobs test cases"""

    def setUp(self):
        super(FailingUserJobTestCase, self).setUp()

        dm = DataManager()
        res = dm.removeFile(
            [
                "/lhcb/testCfg/testVer/LOG/00012345/0006/00012345_00067890.tar",
                "/lhcb/testCfg/testVer/SIM/00012345/0006/00012345_00067890_1.sim",
            ],
            force=True,
        )
        if not res["OK"]:
            print("Could not remove files", res["Message"])
            exit(1)



""" ResourceStatusClient

  Client to interact with the ResourceStatus service and from it with the DB.
"""

# pylint: disable=unused-argument

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import prepareDict


@createClient("ResourceStatus/ResourceStatus")
class ResourceStatusClient(Client):
    """
    The :class:`ResourceStatusClient` class exposes the :mod:`DIRAC.ResourceStatus`
    API. All functions you need are on this client.

    You can use this client on this way

     >>> from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
     >>> rsClient = ResourceStatusClient()
    """

    def __init__(self, **kwargs):

        super(ResourceStatusClient, self).__init__(**kwargs)
        self.setServer("ResourceStatus/ResourceStatus")

    def insert(self, tableName, record):
        """
        Insert a dictionary `record` as a row in table `tableName`

        :param str tableName: the name of the table
        :param dict record: dictionary of record to insert in the table

        :return: S_OK() || S_ERROR()
        """

        return self._getRPC().insert(tableName, record)

    def select(self, tableName, params=None):
        """
        Select rows from the table `tableName`

        :param str tableName: the name of the table
        :param dict record: dictionary of the selection parameters

        :return: S_OK() || S_ERROR()
        """

        if params is None:
            params = {}
        return self._getRPC().select(tableName, params)

    def delete(self, tableName, params=None):
        """
        Delect rows from the table `tableName`

        :param str tableName: the name of the table
        :param dict record: dictionary of the deletion parameters

        :Returns:
          S_OK() || S_ERROR()
        """

        if params is None:
            params = {}
        return self._getRPC().delete(tableName, params)

    ################################################################################
    # Element status methods - enjoy !

    def insertStatusElement(
        self,
        element,
        tableType,
        name,
        statusType,
        status,
        elementType,
        reason,
        dateEffective,
        lastCheckTime,
        tokenOwner,
        tokenExpiration=None,
        vO="all",
    ):
        """
    Inserts on <element><tableType> a new row with the arguments given.

    :Parameters:
      **element** - `string`
        it has to be a valid element ( ValidElement ), any of the defaults: `Site` \
        | `Resource` | `Node`
      **tableType** - `string`
        it has to be a valid tableType [ 'Status', 'Log', 'History' ]
      **name** - `string`
        name of the individual of class element
      **statusType** - `string`
        it has to be a valid status type for the element class
      **status** - `string`
        it has to be a valid status, any of the defaults: `Active` | `Degraded` | \
        `Probing` | `Banned`
      **elementType** - `string`
        column to distinguish between the different elements in the same element
        table.
      **reason** - `string`
        decision that triggered the assigned status
      **dateEffective** - `datetime`
        time-stamp from which the status & status type are effective
      **lastCheckTime** - `datetime`
        time-stamp setting last time the status & status were checked
      **tokenOwner** - `string`
        token assigned to the site & status type
      **tokenExpiration** - `datetime`
        time-stamp setting validity of token ownership

    :return: S_OK() || S_ERROR()
    """
        columnNames = [
            "Name",
            "StatusType",
            "Status",
            "ElementType",
            "Reason",
            "DateEffective",
            "LastCheckTime",
            "TokenOwner",
            "TokenExpiration",
            "VO",
        ]
        columnValues = [
            name,
            statusType,
            status,
            elementType,
            reason,
            dateEffective,
            lastCheckTime,
            tokenOwner,
            tokenExpiration,
            vO,
        ]

        return self._getRPC().insert(element + tableType, prepareDict(columnNames, columnValues))

    def selectStatusElement(
        self,
        element,
        tableType,
        name=None,
        statusType=None,
        status=None,
        elementType=None,
        reason=None,
        dateEffective=None,
        lastCheckTime=None,
        tokenOwner=None,
        tokenExpiration=None,
        meta=None,
        vO="all",
    ):
        """
    Gets from <element><tableType> all rows that match the parameters given.

    :Parameters:
      **element** - `string`
        it has to be a valid element ( ValidElement ), any of the defaults: `Site` \
        | `Resource` | `Node`
      **tableType** - `string`
        it has to be a valid tableType [ 'Status', 'Log', 'History' ]
      **name** - `[, string, list]`
        name of the individual of class element
      **statusType** - `[, string, list]`
        it has to be a valid status type for the element class
      **status** - `[, string, list]`
        it has to be a valid status, any of the defaults: `Active` | `Degraded` | \
        `Probing` | `Banned`
      **elementType** - `[, string, list]`
        column to distinguish between the different elements in the same element
        table.
      **reason** - `[, string, list]`
        decision that triggered the assigned status
      **dateEffective** - `[, datetime, list]`
        time-stamp from which the status & status type are effective
      **lastCheckTime** - `[, datetime, list]`
        time-stamp setting last time the status & status were checked
      **tokenOwner** - `[, string, list]`
        token assigned to the site & status type
      **tokenExpiration** - `[, datetime, list]`
        time-stamp setting validity of token ownership
      **meta** - `dict`
        metadata for the mysql query. Currently it is being used only for column selection.
        For example: meta = { 'columns' : [ 'Name' ] } will return only the 'Name' column.

    :return: S_OK() || S_ERROR()
    """
        columnNames = [
            "Name",
            "StatusType",
            "Status",
            "ElementType",
            "Reason",
            "DateEffective",
            "LastCheckTime",
            "TokenOwner",
            "TokenExpiration",
            "Meta",
            "VO",
        ]
        columnValues = [
            name,
            statusType,
            status,
            elementType,
            reason,
            dateEffective,
            lastCheckTime,
            tokenOwner,
            tokenExpiration,
            meta,
            vO,
        ]

        return self._getRPC().select(element + tableType, prepareDict(columnNames, columnValues))

    def deleteStatusElement(
        self,
        element,
        tableType,
        name=None,
        statusType=None,
        status=None,
        elementType=None,
        reason=None,
        dateEffective=None,
        lastCheckTime=None,
        tokenOwner=None,
        tokenExpiration=None,
        meta=None,
        vO="all",
    ):
        """
    Deletes from <element><tableType> all rows that match the parameters given.

    :Parameters:
      **element** - `string`
        it has to be a valid element ( ValidElement ), any of the defaults: `Site` \
        | `Resource` | `Node`
      **tableType** - `string`
        it has to be a valid tableType [ 'Status', 'Log', 'History' ]
      **name** - `[, string, list]`
        name of the individual of class element
      **statusType** - `[, string, list]`
        it has to be a valid status type for the element class
      **status** - `[, string, list]`
        it has to be a valid status, any of the defaults: `Active` | `Degraded` | \
        `Probing` | `Banned`
      **elementType** - `[, string, list]`
        column to distinguish between the different elements in the same element
        table.
      **reason** - `[, string, list]`
        decision that triggered the assigned status
      **dateEffective** - `[, datetime, list]`
        time-stamp from which the status & status type are effective
      **lastCheckTime** - `[, datetime, list]`
        time-stamp setting last time the status & status were checked
      **tokenOwner** - `[, string, list]`
        token assigned to the site & status type
      **tokenExpiration** - `[, datetime, list]`
        time-stamp setting validity of token ownership
      **meta** - `dict`
        metadata for the mysql query

    :return: S_OK() || S_ERROR()
    """
        columnNames = [
            "Name",
            "StatusType",
            "Status",
            "ElementType",
            "Reason",
            "DateEffective",
            "LastCheckTime",
            "TokenOwner",
            "TokenExpiration",
            "Meta",
            "VO",
        ]
        columnValues = [
            name,
            statusType,
            status,
            elementType,
            reason,
            dateEffective,
            lastCheckTime,
            tokenOwner,
            tokenExpiration,
            meta,
            vO,
        ]

        return self._getRPC().delete(element + tableType, prepareDict(columnNames, columnValues))

    def addOrModifyStatusElement(
        self,
        element,
        tableType,
        name=None,
        statusType=None,
        status=None,
        elementType=None,
        reason=None,
        dateEffective=None,
        lastCheckTime=None,
        tokenOwner=None,
        tokenExpiration=None,
        vO="all",
    ):
        """
    Adds or updates-if-duplicated from <element><tableType> and also adds a log
    if flag is active.

    :Parameters:
      **element** - `string`
        it has to be a valid element ( ValidElement ), any of the defaults: `Site` \
        | `Resource` | `Node`
      **tableType** - `string`
        it has to be a valid tableType [ 'Status', 'Log', 'History' ]
      **name** - `string`
        name of the individual of class element
      **statusType** - `string`
        it has to be a valid status type for the element class
      **status** - `string`
        it has to be a valid status, any of the defaults: `Active` | `Degraded` | \
        `Probing` | `Banned`
      **elementType** - `string`
        column to distinguish between the different elements in the same element
        table.
      **reason** - `string`
        decision that triggered the assigned status
      **dateEffective** - `datetime`
        time-stamp from which the status & status type are effective
      **lastCheckTime** - `datetime`
        time-stamp setting last time the status & status were checked
      **tokenOwner** - `string`
        token assigned to the site & status type
      **tokenExpiration** - `datetime`
        time-stamp setting validity of token ownership

    :return: S_OK() || S_ERROR()
    """
        columnNames = [
            "Name",
            "StatusType",
            "Status",
            "ElementType",
            "Reason",
            "DateEffective",
            "LastCheckTime",
            "TokenOwner",
            "TokenExpiration",
            "VO",
        ]
        columnValues = [
            name,
            statusType,
            status,
            elementType,
            reason,
            dateEffective,
            lastCheckTime,
            tokenOwner,
            tokenExpiration,
            vO,
        ]

        return self._getRPC().addOrModify(element + tableType, prepareDict(columnNames, columnValues))

    def modifyStatusElement(
        self,
        element,
        tableType,
        name=None,
        statusType=None,
        status=None,
        elementType=None,
        reason=None,
        dateEffective=None,
        lastCheckTime=None,
        tokenOwner=None,
        tokenExpiration=None,
        vO="all",
    ):
        """
    Updates from <element><tableType> and also adds a log if flag is active.

    :Parameters:
      **element** - `string`
        it has to be a valid element ( ValidElement ), any of the defaults: `Site` \
        | `Resource` | `Node`
      **tableType** - `string`
        it has to be a valid tableType [ 'Status', 'Log', 'History' ]
      **name** - `string`
        name of the individual of class element
      **statusType** - `string`
        it has to be a valid status type for the element class
      **status** - `string`
        it has to be a valid status, any of the defaults: `Active` | `Degraded` | \
        `Probing` | `Banned`
      **elementType** - `string`
        column to distinguish between the different elements in the same element
        table.
      **reason** - `string`
        decision that triggered the assigned status
      **dateEffective** - `datetime`
        time-stamp from which the status & status type are effective
      **lastCheckTime** - `datetime`
        time-stamp setting last time the status & status were checked
      **tokenOwner** - `string`
        token assigned to the site & status type
      **tokenExpiration** - `datetime`
        time-stamp setting validity of token ownership

    :return: S_OK() || S_ERROR()
    """
        columnNames = [
            "Name",
            "StatusType",
            "Status",
            "ElementType",
            "Reason",
            "DateEffective",
            "LastCheckTime",
            "TokenOwner",
            "TokenExpiration",
            "VO",
        ]
        columnValues = [
            name,
            statusType,
            status,
            elementType,
            reason,
            dateEffective,
            lastCheckTime,
            tokenOwner,
            tokenExpiration,
            vO,
        ]

        return self._getRPC().addOrModify(element + tableType, prepareDict(columnNames, columnValues))

    def addIfNotThereStatusElement(
        self,
        element,
        tableType,
        name=None,
        statusType=None,
        status=None,
        elementType=None,
        reason=None,
        dateEffective=None,
        lastCheckTime=None,
        tokenOwner=None,
        tokenExpiration=None,
        vO="all",
    ):
        """
    Adds if-not-duplicated from <element><tableType> and also adds a log if flag
    is active.

    :Parameters:
      **element** - `string`
        it has to be a valid element ( ValidElement ), any of the defaults: `Site` \
        | `Resource` | `Node`
      **tableType** - `string`
        it has to be a valid tableType [ 'Status', 'Log', 'History' ]
      **name** - `string`
        name of the individual of class element
      **statusType** - `string`
        it has to be a valid status type for the element class
      **status** - `string`
        it has to be a valid status, any of the defaults: `Active` | `Degraded` | \
        `Probing` | `Banned`
      **elementType** - `string`
        column to distinguish between the different elements in the same element
        table.
      **reason** - `string`
        decision that triggered the assigned status
      **dateEffective** - `datetime`
        time-stamp from which the status & status type are effective
      **lastCheckTime** - `datetime`
        time-stamp setting last time the status & status were checked
      **tokenOwner** - `string`
        token assigned to the site & status type
      **tokenExpiration** - `datetime`
        time-stamp setting validity of token ownership

    :return: S_OK() || S_ERROR()
    """
        columnNames = [
            "Name",
            "StatusType",
            "Status",
            "ElementType",
            "Reason",
            "DateEffective",
            "LastCheckTime",
            "TokenOwner",
            "TokenExpiration",
            "VO",
        ]
        columnValues = [
            name,
            statusType,
            status,
            elementType,
            reason,
            dateEffective,
            lastCheckTime,
            tokenOwner,
            tokenExpiration,
            vO,
        ]

        return self._getRPC().addIfNotThere(element + tableType, prepareDict(columnNames, columnValues))

    ##############################################################################
    # Protected methods - Use carefully !!

    def notify(self, request, params):
        """Send notification for a given request with its params to the diracAdmin"""
        address = Operations().getValue("ResourceStatus/Notification/DebugGroup/Users")
        msg = "Matching parameters: " + str(params)
        sbj = "[NOTIFICATION] DIRAC ResourceStatusDB: " + request + " entry"
        NotificationClient().sendMail(address, sbj, msg, address)

    def _extermineStatusElement(self, element, name, keepLogs=True):
        """
    Deletes from <element>Status,
                 <element>History
                 <element>Log
     all rows with `elementName`. It removes all the entries, logs, etc..
    Use with common sense !

    :Parameters:
      **element** - `string`
        it has to be a valid element ( ValidElements ), any of the defaults: \
          `Site` | `Resource` | `Node`
      **name** - `[, string, list]`
        name of the individual of class element
      **keepLogs** - `bool`
        if active, logs are kept in the database

    :return: S_OK() || S_ERROR()
    """
        return self.__extermineStatusElement(element, name, keepLogs)

    def __extermineStatusElement(self, element, name, keepLogs):
        """
        This method iterates over the three ( or four ) table types - depending
        on the value of keepLogs - deleting all matches of `name`.
        """

        tableTypes = ["Status", "History"]
        if keepLogs is False:
            tableTypes.append("Log")

        for table in tableTypes:

            deleteQuery = self.deleteStatusElement(element, table, name=name)
            if not deleteQuery["OK"]:
                return deleteQuery

        return S_OK()


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



"""
This module contains constants and lists for the possible Transformation files states.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__RCSID__ = "$Id$"

#:
UNUSED = "Unused"
#:
ASSIGNED = "Assigned"
#:
PROCESSED = "Processed"
#:
PROBLEMATIC = "Problematic"
#:
MAX_RESET = "MaxReset"
#:
MISSING_IN_FC = "MissingInFC"
#:
PROB_IN_FC = "ProbInFC"
#:
REMOVED = "Removed"

# below ones are for inherited transformations
#:
UNUSED_INHERITED = "Unused-inherited"
#:
ASSIGNED_INHERITED = "Assigned-inherited"
#:
MAXRESET_INHERITED = "MaxReset-inherited"
#:
PROCESSED_INHERITED = "Processed-inherited"
#:
MOVED = "Moved"


#: Possible states
TRANSFORMATION_FILES_STATES = [
    UNUSED,
    ASSIGNED,
    PROCESSED,
    PROBLEMATIC,
    MAX_RESET,
    MISSING_IN_FC,
    PROB_IN_FC,
    UNUSED_INHERITED,
    ASSIGNED_INHERITED,
    MAXRESET_INHERITED,
    PROCESSED_INHERITED,
    MOVED,
]



"""
   Contains the mechanism to evaluate whether to use or not a catalog
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

from pyparsing import infixNotation, opAssoc, Word, printables, Literal, Suppress

from DIRAC import S_OK, gLogger
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations


class FCConditionParser(object):
    """This objects allows to evaluate conditions on whether or not
    a given operation should be evaluated on a given catalog
    for a given lfn (be glad so many things are given to you !).

    The conditions are expressed as boolean logic, where the basic bloc has
    the form "pluginName=whateverThatWillBePassedToThePlugin".
    The basic blocs will be evaluated by the respective plugins, and the result can
    be combined using the standard boolean operators::

      * `!` for not
      * `&` for and
      * `|` for or
      * `[ ]` for prioritizing the operations

    All these characters, as well as the '=' symbol cannot be used in any expression to be
    evaluated by a plugin.

    The rule to evaluate can either be given at calling time, or can be retrieved
    from the CS depending on the context (see doc of __call__ and __getConditionFromCS)


    Example of rules are::

      * Filename=startswith('/lhcb') & Proxy=voms.has(/lhcb/Role->production)
      * [Filename=startswith('/lhcb') & !Filename=find('/user/')] | Proxy=group.in(lhcb_mc, lhcb_data)
    """

    # Some characters are reserved for the grammar
    __forbidenChars = ("[", "]", "!", "&", "|", "=")
    __allowedChars = "".join(set(printables) - set(__forbidenChars)) + " "

    # Defines the basic shape of a rule : pluginName=whateverThatWillBePassedToThePlugin
    __pluginOperand = Word(__allowedChars) + Literal("=") + Word(__allowedChars)

    # define classes to be built at parse time, as each matching
    # expression type is parsed

    # Binary operator base class
    class _BoolBinOp(object):
        """Abstract object to represent a binary operator"""

        reprsymbol = None  # Sign to represent the boolean operation

        # This is the boolean operation to apply
        # Could be None, but it should be callable
        @staticmethod
        def evalop(_x):
            return None

        def __init__(self, token):
            """
            :param token: the token matching a binary operator
                          it is a list with only one element which itself is a list
                          [ [ Arg1, Operator, Arg2] ]
                          The arguments themselves can be of any type, but they need to
                          provide an "eval" method that takes `kwargs` as input,
                          and return a boolean
            """

            # Keep the two arguments
            self.args = token[0][0::2]

        def __str__(self):
            """String representation"""

            sep = " %s " % self.reprsymbol
            return "(" + sep.join(map(str, self.args)) + ")"

        def eval(self, **kwargs):
            """Perform the evaluation of the boolean logic
            by applying the operator between the two arguments

            :param kwargs: whatever information is given to plugin (typically lfn)
            """

            return self.evalop(arg.eval(**kwargs) for arg in self.args)

        __repr__ = __str__

    class _BoolAnd(_BoolBinOp):
        """Represents the 'and' operator"""

        reprsymbol = "&"
        evalop = all

    class _BoolOr(_BoolBinOp):
        """Represents the 'or' operator"""

        reprsymbol = "|"
        evalop = any

    class _BoolNot(object):
        """Represents the "not" unitary operator"""

        def __init__(self, t):
            """
            :param t: the token matching a unitary operator
                      it is a list with only one element which itself is a list
                      [ [ !, Arg1] ]
                      The argument itself can be of any type, but it needs to
                      provide an "eval" method that takes `kwargs` as input,
                      and return a boolean
            """

            # We just keep the argument
            self.arg = t[0][1]

        def eval(self, **kwargs):
            """Perform the evaluation of the boolean logic
            by returning the negation of the evaluation of the argument

            :param kwargs: whatever information is given to plugin (typically lfn)
            """
            return not self.arg.eval(**kwargs)

        def __str__(self):
            return "!" + str(self.arg)

        __repr__ = __str__

    # We can combine the pluginOperand with boolean expression,
    # and prioritized by squared bracket
    __boolExpr = infixNotation(
        __pluginOperand,
        [
            ("!", 1, opAssoc.RIGHT, _BoolNot),
            ("&", 2, opAssoc.LEFT, _BoolAnd),
            ("|", 2, opAssoc.LEFT, _BoolOr),
        ],
        lpar=Suppress("["),
        rpar=Suppress("]"),
    )

    # Wrapper that will call the plugin
    class PluginOperand(object):
        """This class is a wrapper for a plugin
        and it's condition
        It is instantiated by pyparsing every time
        it encounters "plugin=condition"

        """

        def __init__(self, tokens):
            """
            :param tokens: [ pluginName, =, conditions ]
                   the pluginName is automatically prepended with 'Plugin'


            """

            self.pluginName = "%sPlugin" % tokens[0].strip(" ")
            self.conditions = tokens[2].strip(" ")

            # Load the plugin, and give it the condition
            objLoader = ObjectLoader()
            _class = objLoader.loadObject("Resources.Catalog.ConditionPlugins.%s" % self.pluginName)

            if not _class["OK"]:
                raise Exception(_class["Message"])

            self._pluginInst = _class["Value"](self.conditions)

        def eval(self, **kwargs):
            """Forward the evaluation call to the plugin

            :param kwargs: contains all the information given to the plugin namely the lfns
            :return: True or False
            """

            return self._pluginInst.eval(**kwargs)

        def __str__(self):
            return self.pluginName

        __repr__ = __str__

    def __init__(self, vo=None, ro_methods=None):
        """
        :param vo: name of the VO
        """

        # Whenever we parse text matching the __pluginOperand grammar, create a PluginOperand object
        self.__pluginOperand.setParseAction(lambda tokens: self.PluginOperand(tokens))

        self.opHelper = Operations(vo=vo)

        self.ro_methods = ro_methods if ro_methods else []

        self.log = gLogger.getSubLogger("FCConditionParser")

    def __evaluateCondition(self, conditionString, **kwargs):
        """Evaluate a condition against attributes, typically lfn.
        CAUTION: lfns are here given one by one

        """

        self.log.debug("Testing %s against %s" % (conditionString, kwargs))

        # Parse all the condition and evaluate it
        # res is a tuple whose first and only element is either
        # one of the bool operator defined above, or a PluginOperand
        res = self.__boolExpr.parseString(conditionString)
        res = res[0].eval(**kwargs)

        self.log.debug("Evaluated to %s" % res)

        return res

    def __getConditionFromCS(self, catalogName, operationName):
        """Retrieves the appropriate condition from the CS
        The base path is in Operation/[Setup/Default]/DataManagement/FCConditions/[CatalogName]
        If there are no condition defined for the method, we check the global READ/WRITE condition.
        If this does not exist either, we check the global ALL condition.
        If none is defined, we return None

        :param str catalogName: the catalog we want to work on
        :param str operationName: the operation we want to perform
                                  The operationName must be in the read or write method from FileCatalog

        :returns: a condition string or None


        """
        basePath = "Services/Catalogs/%s/Conditions/" % catalogName
        pathList = [
            basePath + "%s" % operationName,
            basePath + "%s" % ("READ" if operationName in self.ro_methods else "WRITE"),
            basePath + "ALL",
        ]

        for path in pathList:
            condVal = self.opHelper.getValue(path)
            if condVal:
                return condVal

    def __call__(self, catalogName, operationName, lfns, condition=None, **kwargs):
        """
        Makes a boolean evaluation of a condition, for a given catalog,
        a given operation, and a list of lfns. Extra parameters might be given,
        and will be forwarded to each plugin.
        If the 'condition' attribute is not specified (general case), it is fetched from the CS
        (see __getConditionFromCS)

        If there are no condition at all, return True for everything.
        A programming error in the plugins will lead to the evaluation being False

        .. Note::

          if the CS can't be contacted, the conditions will be evaluated to None
          (courtesy of the Operation helper), so everything will be evaluated to True.
          Ultimately, it does not really matter, since you will not be able to find
          any catalog beforehand if you can't contact the CS...

        :param str catalogName: name of the catalog we want to work on
        :param str operationName: name of the operation we want to perform
                              The operationName must be in the read or write method from FileCatalog
                              if it should be retrieve from the CS
        :param lfns: list/dict of lfns

          .. warning::

             LFNs are expected to have been through the normalizing process, so it
             should not be a string

        :param condition: condition string. If not specified, will be fetched from the CS
        :param kwargs: extra params forwarded to the plugins

        :return: S_OK with a 'Successful' dict {lfn:True/False} where the value is the evaluation of the
                condition against the given lfn key. Failed dict is always empty


        """
        self.log.debug("Testing %s on %s for %s lfns" % (operationName, catalogName, len(lfns)))

        conditionStr = condition if condition is not None else self.__getConditionFromCS(catalogName, operationName)

        self.log.debug("Condition string: %s" % conditionStr)

        evaluatedLfns = {}

        if conditionStr:
            for lfn in lfns:
                try:
                    evaluatedLfns[lfn] = self.__evaluateCondition(conditionStr, lfn=lfn, **kwargs)
                except Exception as e:
                    self.log.exception("Exception while evaluation conditions", lException=e)
                    evaluatedLfns[lfn] = False
        else:
            evaluatedLfns = dict.fromkeys(lfns, True)

        return S_OK({"Successful": evaluatedLfns, "Failed": {}})



"""
This class is used to define the plot using the plot attributes.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK

from DIRAC.MonitoringSystem.Client.Types.WMSHistory import WMSHistory
from DIRAC.MonitoringSystem.private.Plotters.BasePlotter import BasePlotter

__RCSID__ = "$Id$"


class WMSHistoryPlotter(BasePlotter):

    """
    .. class:: WMSHistoryPlotter

    It is used to crate the plots.

    param: str _typeName monitoring type
    param: list _typeKeyFields list of keys what we monitor (list of attributes)
    """

    _typeName = "WMSHistory"
    _typeKeyFields = WMSHistory().keyFields

    def _reportNumberOfJobs(self, reportRequest):
        """It is used to retrieve the data from the database.

        :param dict reportRequest: contains attributes used to create the plot.
        :return: S_OK or S_ERROR {'data':value1, 'granularity':value2} value1 is a dictionary, value2 is the bucket length
        """
        retVal = self._getTimedData(
            startTime=reportRequest["startTime"],
            endTime=reportRequest["endTime"],
            selectField="Jobs",
            preCondDict=reportRequest["condDict"],
            metadataDict=None,
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotNumberOfJobs(self, reportRequest, plotInfo, filename):
        """It creates the plot.

        :param dict reportRequest: plot attributes
        :param dict plotInfo: contains all the data which are used to create the plot
        :param str filename:
        :return: S_OK or S_ERROR { 'plot' : value1, 'thumbnail' : value2 } value1 and value2 are TRUE/FALSE
        """
        metadata = {
            "title": "Jobs by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "skipEdgeColor": True,
            "ylabel": "jobs",
        }

        plotInfo["data"] = self._fillWithZero(
            granularity=plotInfo["granularity"],
            startEpoch=reportRequest["startTime"],
            endEpoch=reportRequest["endTime"],
            dataDict=plotInfo["data"],
        )

        return self._generateStackedLinePlot(filename=filename, dataDict=plotInfo["data"], metadata=metadata)

    def _reportNumberOfReschedules(self, reportRequest):
        """It is used to retrieve the data from the database.

        :param dict reportRequest: contains attributes used to create the plot.
        :return: S_OK or S_ERROR {'data':value1, 'granularity':value2} value1 is a dictionary, value2 is the bucket length
        """
        retVal = self._getTimedData(
            startTime=reportRequest["startTime"],
            endTime=reportRequest["endTime"],
            selectField="Reschedules",
            preCondDict=reportRequest["condDict"],
            metadataDict=None,
        )

        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotNumberOfReschedules(self, reportRequest, plotInfo, filename):
        """It creates the plot.

        :param dict reportRequest: plot attributes
        :param dict plotInfo: contains all the data which are used to create the plot
        :param str filename:
        :return S_OK or S_ERROR { 'plot' : value1, 'thumbnail' : value2 } value1 and value2 are TRUE/FALSE
        """
        metadata = {
            "title": "Reschedules by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "skipEdgeColor": True,
            "ylabel": "reschedules",
        }

        plotInfo["data"] = self._fillWithZero(
            granularity=plotInfo["granularity"],
            startEpoch=reportRequest["startTime"],
            endEpoch=reportRequest["endTime"],
            dataDict=plotInfo["data"],
        )

        return self._generateStackedLinePlot(filename=filename, dataDict=plotInfo["data"], metadata=metadata)

    def _reportAverageNumberOfJobs(self, reportRequest):
        """It is used to retrieve the data from the database.

        :param dict reportRequest: contains attributes used to create the plot.
        :return: S_OK or S_ERROR {'data':value1, 'granularity':value2} value1 is a dictionary, value2 is the bucket length
        """
        retVal = self._getSummaryData(
            startTime=reportRequest["startTime"],
            endTime=reportRequest["endTime"],
            selectField="Jobs",
            preCondDict=reportRequest["condDict"],
            metadataDict={"metric": "avg"},
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        return S_OK({"data": dataDict})

    def _plotAverageNumberOfJobs(self, reportRequest, plotInfo, filename):
        """It creates the plot.

        :param dict reportRequest: plot attributes
        :param dict plotInfo: contains all the data which are used to create the plot
        :param str filename:
        :return S_OK or S_ERROR { 'plot' : value1, 'thumbnail' : value2 } value1 and value2 are TRUE/FALSE
        """
        metadata = {
            "title": "Average Number of Jobs by %s" % reportRequest["grouping"],
            "ylabel": "Jobs",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }

        return self._generatePiePlot(filename=filename, dataDict=plotInfo["data"], metadata=metadata)



""" DIRAC FileCatalog client utilities
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os
import errno
import functools

from DIRAC import S_OK, S_ERROR


def checkArgumentFormat(path, generateMap=False):
    """Bring the various possible form of arguments to FileCatalog methods to
    the standard dictionary form
    """

    def checkArgumentDict(path):
        """Check and process format of the arguments to FileCatalog methods"""
        if isinstance(path, six.string_types):
            urls = {path: True}
        elif isinstance(path, list):
            urls = {}
            for url in path:
                urls[url] = True
        elif isinstance(path, dict):
            urls = path
        else:
            return S_ERROR(errno.EINVAL, "Utils.checkArgumentFormat: Supplied path is not of the correct format.")
        return S_OK(urls)

    if not path:
        return S_ERROR(errno.EINVAL, "Empty input: %s" % str(path))

    result = checkArgumentDict(path)
    if not result["OK"]:
        return result

    pathDict = result["Value"]

    # Bring the lfn path to the normalized form
    urls = {}
    urlMap = {}
    for url in pathDict:
        # avoid empty path...
        if not url:
            continue
        mUrl = url
        if url.lower().startswith("lfn:"):
            mUrl = url[4:]
        # Strip off the leading /grid prefix as required for the LFC
        if mUrl.startswith("/grid/"):
            uList = mUrl.split("/")
            uList.pop(1)
            mUrl = "/".join(uList)
        normPath = os.path.normpath(mUrl)
        urls[normPath] = pathDict[url]
        if normPath != url:
            urlMap[normPath] = url
    if generateMap:
        return S_OK((urls, urlMap))
    else:
        return S_OK(urls)


def checkCatalogArguments(f):
    """Decorator to check arguments of FileCatalog calls in the clients"""

    @functools.wraps(f)
    def processWithCheckingArguments(*args, **kwargs):

        checkFlag = kwargs.pop("LFNChecking", True)
        if checkFlag:
            argList = list(args)
            lfnArgument = argList[1]
            result = checkArgumentFormat(lfnArgument, generateMap=True)
            if not result["OK"]:
                return result
            checkedLFNDict, lfnMap = result["Value"]
            argList[1] = checkedLFNDict
            argTuple = tuple(argList)
        else:
            argTuple = args
        result = f(*argTuple, **kwargs)
        if not result["OK"]:
            return result

        if not checkFlag:
            return result

        # Restore original paths
        argList[1] = lfnArgument
        failed = {}
        successful = {}
        for lfn in result["Value"]["Failed"]:
            failed[lfnMap.get(lfn, lfn)] = result["Value"]["Failed"][lfn]
        for lfn in result["Value"]["Successful"]:
            successful[lfnMap.get(lfn, lfn)] = result["Value"]["Successful"][lfn]

        result["Value"].update({"Successful": successful, "Failed": failed})
        return result

    return processWithCheckingArguments



#!/usr/bin/env python
"""
Allows to add a specified extension to an already existing DIRAC installation.
The extension can come from another project than the one installed.
No new version directory is created. The command is based on the main DIRAC installer dirac-install.py.

Usage:
  dirac-install-extension [options] ...

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import sys
import six
import getopt
import importlib

cmdOpts = (
    ("r:", "release=", "Release version to install"),
    ("l:", "project=", "Project to install"),
    ("e:", "extensions=", "Extensions to install (comma separated). Several -e options can be given"),
    ("M:", "defaultsURL=", "Where to retrieve the global defaults from"),
    ("h", "help", "help doc string"),
)


def usage():
    """Usage printout"""
    print(__doc__)
    print("Options::\n\n")
    for cmdOpt in cmdOpts:
        print("  %s %s : %s" % (cmdOpt[0].ljust(3), cmdOpt[1].ljust(20), cmdOpt[2]))

    sys.exit(0)


# Import dirac-install.py as a module
installFile = ""
for basePath in ("pro", "."):
    if os.path.exists(os.path.join(basePath, "DIRAC/Core/scripts/dirac-install.py")):
        installFile = os.path.join(basePath, "DIRAC/Core/scripts/dirac-install.py")
        break

if installFile:
    sys.path.append(os.path.dirname(installFile))
    diracInstall = importlib.import_module("dirac-install")
else:
    usage()
    sys.exit(-1)


def loadConfiguration():
    """
    It loads the configuration file
    """
    optList, args = getopt.getopt(sys.argv[1:], "".join([opt[0] for opt in cmdOpts]), [opt[1] for opt in cmdOpts])

    # First check if the name is defined
    for opt, value in optList:
        if opt in ("-h", "--help"):
            usage()
        elif opt in ("-M", "--defaultsURL"):
            diracInstall.cliParams.globalDefaults = value

    rConfig = diracInstall.ReleaseConfig(
        instName=diracInstall.cliParams.installation, globalDefaultsURL=diracInstall.cliParams.globalDefaults
    )
    if diracInstall.cliParams.debug:
        rConfig.debugCB = diracInstall.logDEBUG

    res = rConfig.loadInstallationDefaults()
    if not res["OK"]:
        diracInstall.logERROR("Could not load defaults: %s" % res["Message"])

    rConfig.loadInstallationLocalDefaults(args)

    for opName in ("release", "globalDefaults", "useVersionsDir"):
        try:
            opVal = rConfig.getInstallationConfig("LocalInstallation/%s" % (opName[0].upper() + opName[1:]))
        except KeyError:
            continue

        if isinstance(getattr(diracInstall.cliParams, opName), six.string_types):
            setattr(diracInstall.cliParams, opName, opVal)
        elif isinstance(getattr(diracInstall.cliParams, opName), bool):
            setattr(diracInstall.cliParams, opName, opVal.lower() in ("y", "yes", "true", "1"))

    # Now parse the ops
    for opt, value in optList:
        if opt in ("-r", "--release"):
            diracInstall.cliParams.release = value
        elif opt in ("-l", "--project"):
            diracInstall.cliParams.project = value
        elif opt in ("-e", "--extensions"):
            for pkg in [p.strip() for p in value.split(",") if p.strip()]:
                if pkg not in diracInstall.cliParams.extensions:
                    diracInstall.cliParams.extensions.append(pkg)

    if not diracInstall.cliParams.release and not diracInstall.cliParams.modules:
        diracInstall.logERROR("Missing release to install")
        usage()

    diracInstall.cliParams.basePath = diracInstall.cliParams.targetPath
    if diracInstall.cliParams.useVersionsDir:
        # install under the pro directory
        diracInstall.cliParams.targetPath = os.path.join(diracInstall.cliParams.targetPath, "pro")

    diracInstall.logNOTICE("Destination path for installation is %s" % diracInstall.cliParams.targetPath)
    rConfig.projectName = diracInstall.cliParams.project

    res = rConfig.loadProjectRelease(
        diracInstall.cliParams.release,
        project=diracInstall.cliParams.project,
        sourceURL=diracInstall.cliParams.installSource,
    )
    if not res["OK"]:
        return res

    # Reload the local configuration to ensure it takes prescience
    rConfig.loadInstallationLocalDefaults(args)

    return diracInstall.S_OK(rConfig)


if __name__ == "__main__":

    result = loadConfiguration()
    if result["OK"]:
        releaseConfig = result["Value"]
    else:
        diracInstall.logERROR("Can not load the configuration: %s" % result["Message"])
        sys.exit(-1)

    result = releaseConfig.getModulesToInstall(diracInstall.cliParams.release, diracInstall.cliParams.extensions)

    for extension in diracInstall.cliParams.extensions:
        if ":" in extension:
            extension = extension.split(":")[0]
        extUrl = result["Value"][1][extension][0]
        extVersion = result["Value"][1][extension][1]

        diracInstall.logNOTICE("Installing extension %s:%s" % (extension, extVersion))

        if not diracInstall.downloadAndExtractTarball(extUrl, extension, extVersion):
            diracInstall.logERROR("Failed to install %s" % extension)
            sys.exit(-1)

    # (Re)deploy scripts now taking into account the newly installed extensions
    ddeLocation = os.path.join(diracInstall.cliParams.targetPath, "DIRAC", "Core", "scripts", "dirac-deploy-scripts.py")
    if not os.path.isfile(ddeLocation):
        ddeLocation = os.path.join(
            diracInstall.cliParams.targetPath, "DIRAC", "Core", "scripts", "dirac_deploy_scripts.py"
        )
    if os.path.isfile(ddeLocation):
        cmd = ddeLocation

    # if specified, create symlink instead of wrapper.
    if diracInstall.cliParams.scriptSymlink:
        cmd += " --symlink"

    extensionList = []
    for extension in diracInstall.cliParams.extensions:
        if ":" in extension:
            extension = extension.split(":")[0]
        extensionList.append(extension.split(":")[0])
    cmd += " --module " + ",".join(extensionList)
    os.system(cmd)



""" Collection of utilities for locating certs, proxy, CAs
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import DIRAC
from DIRAC import gConfig

g_SecurityConfPath = "/DIRAC/Security"


def getProxyLocation():
    """Get the path of the currently active grid proxy file"""

    for envVar in ["GRID_PROXY_FILE", "X509_USER_PROXY"]:
        if envVar in os.environ:
            proxyPath = os.path.realpath(os.environ[envVar])
            if os.path.isfile(proxyPath):
                return proxyPath
    # /tmp/x509up_u<uid>
    proxyName = "x509up_u%d" % os.getuid()
    if os.path.isfile("/tmp/%s" % proxyName):
        return "/tmp/%s" % proxyName

    # No gridproxy found
    return False


# Retrieve CA's location


def getCAsLocation():
    """Retrieve the CA's files location"""
    # Grid-Security
    retVal = gConfig.getOption("%s/Grid-Security" % g_SecurityConfPath)
    if retVal["OK"]:
        casPath = "%s/certificates" % retVal["Value"]
        if os.path.isdir(casPath):
            return casPath
    # CAPath
    retVal = gConfig.getOption("%s/CALocation" % g_SecurityConfPath)
    if retVal["OK"]:
        casPath = retVal["Value"]
        if os.path.isdir(casPath):
            return casPath
    # Look up the X509_CERT_DIR environment variable
    if "X509_CERT_DIR" in os.environ:
        casPath = os.environ["X509_CERT_DIR"]
        return casPath
    # rootPath./etc/grid-security/certificates
    casPath = "%s/etc/grid-security/certificates" % DIRAC.rootPath
    if os.path.isdir(casPath):
        return casPath
    # /etc/grid-security/certificates
    casPath = "/etc/grid-security/certificates"
    if os.path.isdir(casPath):
        return casPath
    # No CA's location found
    return False


# Retrieve CA's location


def getCAsDefaultLocation():
    """Retrievethe CAs Location inside DIRAC etc directory"""
    # rootPath./etc/grid-security/certificates
    casPath = "%s/etc/grid-security/certificates" % DIRAC.rootPath
    return casPath


# TODO: Static depending on files specified on CS
# Retrieve certificate


def getHostCertificateAndKeyLocation(specificLocation=None):
    """Retrieve the host certificate files location.

    Lookup order:

    * ``specificLocation`` (probably broken, don't use it)
    * Environment variables (``DIRAC_X509_HOST_CERT`` and ``DIRAC_X509_HOST_KEY``)
    * CS (``/DIRAC/Security/CertFile`` and ``/DIRAC/Security/CertKey``)
    * Alternative exotic options, with ``prefix`` in  ``server``, ``host``, ``dirac``, ``service``:
      * in `<DIRAC rootpath>/etc/grid-security/` for ``<prefix>cert.pem`` and ``<prefix>key.pem``
      * in the path defined in the CS in ``/DIRAC/Security/Grid-Security``

    :param specificLocation: CS path to look for a the path to cert and key, which then should be the same.
                             Probably does not work, don't use it

    :returns: tuple ``(<cert location>, <key location>)`` or ``False``

    """

    fileDict = {}

    # First, check the environment variables
    for fileType, envVar in (("cert", "DIRAC_X509_HOST_CERT"), ("key", "DIRAC_X509_HOST_KEY")):
        if envVar in os.environ and os.path.exists(os.environ[envVar]):
            fileDict[fileType] = os.environ[envVar]

    for fileType in ("cert", "key"):
        # Check if we already have the info
        if fileType in fileDict:
            continue

        # Direct file in config
        retVal = gConfig.getOption("%s/%sFile" % (g_SecurityConfPath, fileType.capitalize()))
        if retVal["OK"]:
            fileDict[fileType] = retVal["Value"]
            continue
        fileFound = False
        for filePrefix in ("server", "host", "dirac", "service"):
            # Possible grid-security's
            paths = []
            retVal = gConfig.getOption("%s/Grid-Security" % g_SecurityConfPath)
            if retVal["OK"]:
                paths.append(retVal["Value"])
            paths.append("%s/etc/grid-security/" % DIRAC.rootPath)
            for path in paths:
                filePath = os.path.realpath("%s/%s%s.pem" % (path, filePrefix, fileType))
                if os.path.isfile(filePath):
                    fileDict[fileType] = filePath
                    fileFound = True
                    break
            if fileFound:
                break
    if "cert" not in fileDict or "key" not in fileDict:
        return False
    # we can specify a location outside /opt/dirac/etc/grid-security directory
    if specificLocation:
        fileDict["cert"] = gConfig.getValue(specificLocation, fileDict["cert"])
        fileDict["key"] = gConfig.getValue(specificLocation, fileDict["key"])

    return (fileDict["cert"], fileDict["key"])


def getCertificateAndKeyLocation():
    """Get the locations of the user X509 certificate and key pem files"""

    certfile = ""
    if "X509_USER_CERT" in os.environ:
        if os.path.exists(os.environ["X509_USER_CERT"]):
            certfile = os.environ["X509_USER_CERT"]
    if not certfile:
        if os.path.exists(os.environ["HOME"] + "/.globus/usercert.pem"):
            certfile = os.environ["HOME"] + "/.globus/usercert.pem"

    if not certfile:
        return False

    keyfile = ""
    if "X509_USER_KEY" in os.environ:
        if os.path.exists(os.environ["X509_USER_KEY"]):
            keyfile = os.environ["X509_USER_KEY"]
    if not keyfile:
        if os.path.exists(os.environ["HOME"] + "/.globus/userkey.pem"):
            keyfile = os.environ["HOME"] + "/.globus/userkey.pem"

    if not keyfile:
        return False

    return (certfile, keyfile)


def getDefaultProxyLocation():
    """Get the location of a possible new grid proxy file"""

    for envVar in ["GRID_PROXY_FILE", "X509_USER_PROXY"]:
        if envVar in os.environ:
            proxyPath = os.path.realpath(os.environ[envVar])
            return proxyPath

    # /tmp/x509up_u<uid>
    proxyName = "x509up_u%d" % os.getuid()
    return "/tmp/%s" % proxyName



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-attributes
# Author :  Stuart Paterson
########################################################################
"""
Retrieve attributes associated with the given DIRAC job

Example:
  $ dirac-wms-job-attributes  1
  {'AccountedFlag': 'False',
   'ApplicationNumStatus': '0',
   'ApplicationStatus': 'Unknown',
   'CPUTime': '0.0',
   'DIRACSetup': 'EELA-Production',
   'DeletedFlag': 'False',
   'EndExecTime': '2011-02-14 11:28:01',
   'FailedFlag': 'False',
   'HeartBeatTime': '2011-02-14 11:28:01',
   'ISandboxReadyFlag': 'False',
   'JobGroup': 'NoGroup',
   'JobID': '1',
   'JobName': 'DIRAC_vhamar_602138',
   'JobSplitType': 'Single',
   'JobType': 'normal',
   'KilledFlag': 'False',
   'LastUpdateTime': '2011-02-14 11:28:11',
   'MasterJobID': '0',
   'MinorStatus': 'Execution Complete',
   'OSandboxReadyFlag': 'False',
   'Owner': 'vhamar',
   'OwnerDN': '/O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar',
   'OwnerGroup': 'eela_user',
   'RescheduleCounter': '0',
   'RescheduleTime': 'None',
   'RetrievedFlag': 'False',
   'RunNumber': '0',
   'Site': 'EELA.UTFSM.cl',
   'StartExecTime': '2011-02-14 11:27:48',
   'Status': 'Done',
   'SubmissionTime': '2011-02-14 10:12:40',
   'SystemPriority': '0',
   'UserPriority': '1',
   'VerifiedFlag': 'True'}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments

    dirac = Dirac()
    exitCode = 0
    errorList = []

    for job in parseArguments(args):

        result = dirac.getJobAttributes(job, printOutput=True)
        if not result["OK"]:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-kill
# Author :  Stuart Paterson
########################################################################
"""
Issue a kill signal to a running DIRAC job

Example:
  $ dirac-wms-job-kill 1918
  Killed job 1918

.. Note::

  - jobs will not disappear from JobDB until JobCleaningAgent has deleted them
  - jobs will be deleted "immediately" if they are in the status 'Deleted'
  - USER jobs will be deleted after a grace period if they are in status Killed, Failed, Done

  What happens when you hit the "kill job" button

  - if the job is in status 'Running', 'Matched', 'Stalled' it will be properly killed, and then its
    status will be marked as 'Killed'
  - otherwise, it will be marked directly as 'Killed'.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments

    result = Dirac().killJob(parseArguments(args))
    if result["OK"]:
        print("Killed jobs %s" % ",".join([str(j) for j in result["Value"]]))
        exitCode = 0
    else:
        print("ERROR", result["Message"])
        exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac_proxy_get_uploaded_info.py
# Author :  Adrian Casajus
########################################################################
"""
Print information about uploaded proxies.

Example:
  $ dirac-proxy-get-uploaded-info
  Checking for DNs /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar
  --------------------------------------------------------------------------------------------------------
  | UserDN                                          | UserGroup   | ExpirationTime      | PersistentFlag |
  --------------------------------------------------------------------------------------------------------
  | /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar | dirac_user  | 2011-06-29 12:04:25 | True           |
  --------------------------------------------------------------------------------------------------------
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import sys

from DIRAC import gLogger, S_OK
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import ProxyManagerClient
from DIRAC.Core.Security import Properties
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.ConfigurationSystem.Client.Helpers import Registry

userName = False


def setUser(arg):
    """Set user

    :param str arg: user name

    :return: S_OK()
    """
    global userName
    userName = arg
    return S_OK()


@Script()
def main():
    global userName
    Script.registerSwitch("u:", "user=", "User to query (by default oneself)", setUser)
    Script.parseCommandLine()

    result = getProxyInfo()
    if not result["OK"]:
        gLogger.notice("Do you have a valid proxy?")
        gLogger.notice(result["Message"])
        sys.exit(1)
    proxyProps = result["Value"]

    userName = userName or proxyProps.get("username")
    if not userName:
        gLogger.notice("Your proxy don`t have username extension")
        sys.exit(1)

    if userName in Registry.getAllUsers():
        if Properties.PROXY_MANAGEMENT not in proxyProps["groupProperties"]:
            if userName != proxyProps["username"] and userName != proxyProps["issuer"]:
                gLogger.notice("You can only query info about yourself!")
                sys.exit(1)
        result = Registry.getDNForUsername(userName)
        if not result["OK"]:
            gLogger.notice("Oops %s" % result["Message"])
        dnList = result["Value"]
        if not dnList:
            gLogger.notice("User %s has no DN defined!" % userName)
            sys.exit(1)
        userDNs = dnList
    else:
        userDNs = [userName]

    gLogger.notice("Checking for DNs %s" % " | ".join(userDNs))
    pmc = ProxyManagerClient()
    result = pmc.getDBContents({"UserDN": userDNs})
    if not result["OK"]:
        gLogger.notice("Could not retrieve the proxy list: %s" % result["Value"])
        sys.exit(1)

    data = result["Value"]
    colLengths = []
    for pN in data["ParameterNames"]:
        colLengths.append(len(pN))
    for row in data["Records"]:
        for i in range(len(row)):
            colLengths[i] = max(colLengths[i], len(str(row[i])))

    lines = [""]
    for i in range(len(data["ParameterNames"])):
        pN = data["ParameterNames"][i]
        lines[0] += "| %s " % pN.ljust(colLengths[i])
    lines[0] += "|"
    tL = len(lines[0])
    lines.insert(0, "-" * tL)
    lines.append("-" * tL)
    for row in data["Records"]:
        nL = ""
        for i in range(len(row)):
            nL += "| %s " % str(row[i]).ljust(colLengths[i])
        nL += "|"
        lines.append(nL)
        lines.append("-" * tL)

    gLogger.notice("\n".join(lines))


if __name__ == "__main__":
    main()



"""
  Defines the plugin to return the boolean given as param
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

from DIRAC.Resources.Catalog.ConditionPlugins.FCConditionBasePlugin import FCConditionBasePlugin


class DummyPlugin(FCConditionBasePlugin):
    """
    This plugin is to be used to simply return True or False
    """

    def __init__(self, conditions):
        """the condition can be True or False:"""
        super(DummyPlugin, self).__init__(conditions)

    def eval(self, **kwargs):
        """evaluate whether the conditon is True or False"""
        return eval(self.conditions)



########################################################################
# File: OperationHandlerBase.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/03/13 13:48:52
########################################################################
""" :mod: OperationHandlerBase

    ==========================

    .. module: OperationHandlerBase

    :synopsis: request operation handler base class

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    RMS Operation handler base class.

    This should be a functor getting Operation as ctor argument and calling it (executing __call__)
    should return S_OK/S_ERROR.

    Helper functions and tools:

    * self.rssClient() -- returns RSSClient
    * self.getProxyForLFN( LFN ) -- sets X509_USER_PROXY environment variable to LFN owner proxy
    * self.rssSEStatus( SE, status ) returns S_OK(True/False) depending of RSS :status:

    Properties:

    * self.shifter -- list of shifters matching request owner (could be empty!!!)
    * each CS option stored under CS path "RequestExecutingAgent/OperationHandlers/Foo" is exported as read-only property too
    * self.initialize() -- overwrite it to perform additional initialization
    * self.log -- own sub logger
    * self.request, self.operation -- reference to Operation and Request itself

    In all inherited class one should overwrite __call__ and initialize, when appropriate.

    For monitoring purpose each of operation handler has got defined at this level three
    :gMonitor: activities to be used together with given operation.Type, namely
    operation.Type + "Att", operation.Type + "Succ" and operation.Type + "Fail", i.e. for
    operation.Type = "Foo", they are "FooAtt", "FooSucc", "FooFail". Treating of those is done
    automatically, but if you need to monitor more, DIY.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"
# #
# @file OperationHandlerBase.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/13 13:49:02
# @brief Definition of OperationHandlerBase class.

# # imports
import os
import six

# # from DIRAC
from DIRAC import gLogger, gConfig, S_ERROR, S_OK
from DIRAC.Core.Utilities.Graph import DynamicProps
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getGroupsWithVOMSAttribute
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Core.Utilities import Time, Network


@six.add_metaclass(DynamicProps)
class OperationHandlerBase(object):
    """
    .. class:: OperationHandlerBase

    request operation handler base class
    """

    # # private data logging client
    #   __dataLoggingClient = None
    # # private ResourceStatusClient
    __rssClient = None
    # # shifter list
    __shifterList = []

    def __init__(self, operation=None, csPath=None):
        """c'tor

        :param Operation operation: Operation instance
        :param str csPath: config path in CS for this operation
        """
        # # placeholders for operation and request
        self.operation = None
        self.request = None

        self.rmsMonitoring = False

        self.dm = DataManager()
        self.fc = FileCatalog()

        self.csPath = csPath if csPath else ""
        # # get name
        name = self.__class__.__name__
        # # all options are r/o properties now
        csOptionsDict = gConfig.getOptionsDict(self.csPath)
        csOptionsDict = csOptionsDict.get("Value", {})

        for option, value in csOptionsDict.items():
            # # hack to set proper types
            try:
                value = eval(value)
            except NameError:
                pass
            self.makeProperty(option, value, True)  # pylint: disable=no-member

        # # pre setup logger
        self.log = gLogger.getSubLogger(name)
        # # set log level
        logLevel = getattr(self, "LogLevel") if hasattr(self, "LogLevel") else "INFO"
        self.log.setLevel(logLevel)

        # # list properties
        for option in csOptionsDict:
            self.log.debug("%s = %s" % (option, getattr(self, option)))

        # # setup operation
        if operation:
            self.setOperation(operation)
        # # initialize at least
        if hasattr(self, "initialize") and callable(getattr(self, "initialize")):
            getattr(self, "initialize")()

    def setOperation(self, operation):
        """operation and request setter

        :param ~DIRAC.RequestManagementSystem.Client.Operation.Operation operation: operation instance
        :raises TypeError: if `operation` in not an instance of :class:`~DIRAC.RequestManagementSystem.Client.Operation.Operation`

        """
        if not isinstance(operation, Operation):
            raise TypeError("expecting Operation instance")
        self.operation = operation
        self.request = operation._parent
        self.log = gLogger.getSubLogger(
            "pid_%s/%s/%s/%s" % (os.getpid(), self.request.RequestName, self.request.Order, self.operation.Type)
        )

    #   @classmethod
    #   def dataLoggingClient( cls ):
    #     """ DataLoggingClient getter """
    #     if not cls.__dataLoggingClient:
    #       from DIRAC.DataManagementSystem.Client.DataLoggingClient import DataLoggingClient
    #       cls.__dataLoggingClient = DataLoggingClient()
    #     return cls.__dataLoggingClient

    @classmethod
    def rssClient(cls):
        """ResourceStatusClient getter"""
        if not cls.__rssClient:
            from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus

            cls.__rssClient = ResourceStatus()
        return cls.__rssClient

    def getProxyForLFN(self, lfn):
        """get proxy for lfn

        :param str lfn: LFN
        :return: S_ERROR or S_OK( "/path/to/proxy/file" )
        """
        dirMeta = returnSingleResult(self.fc.getDirectoryMetadata(os.path.dirname(lfn)))
        if not dirMeta["OK"]:
            return dirMeta
        dirMeta = dirMeta["Value"]

        ownerRole = "/%s" % dirMeta["OwnerRole"] if not dirMeta["OwnerRole"].startswith("/") else dirMeta["OwnerRole"]
        ownerDN = dirMeta["OwnerDN"]

        ownerProxy = None
        for ownerGroup in getGroupsWithVOMSAttribute(ownerRole):
            vomsProxy = gProxyManager.downloadVOMSProxy(
                ownerDN, ownerGroup, limited=True, requiredVOMSAttribute=ownerRole
            )
            if not vomsProxy["OK"]:
                self.log.debug(
                    "getProxyForLFN: failed to get VOMS proxy for %s role=%s: %s"
                    % (ownerDN, ownerRole, vomsProxy["Message"])
                )
                continue
            ownerProxy = vomsProxy["Value"]
            self.log.debug("getProxyForLFN: got proxy for %s@%s [%s]" % (ownerDN, ownerGroup, ownerRole))
            break

        if not ownerProxy:
            return S_ERROR("Unable to get owner proxy")

        dumpToFile = ownerProxy.dumpAllToFile()
        if not dumpToFile["OK"]:
            self.log.debug("getProxyForLFN: error dumping proxy to file: %s" % dumpToFile["Message"])
        else:
            os.environ["X509_USER_PROXY"] = dumpToFile["Value"]
        return dumpToFile

    def getWaitingFilesList(self):
        """prepare waiting files list, update Attempt, filter out MaxAttempt"""
        if not self.operation:
            self.log.warning("getWaitingFilesList: operation not set, returning empty list")
            return []
        waitingFiles = [opFile for opFile in self.operation if opFile.Status == "Waiting"]
        for opFile in waitingFiles:
            opFile.Attempt += 1
            maxAttempts = getattr(self, "MaxAttempts") if hasattr(self, "MaxAttempts") else 1024
            if opFile.Attempt > maxAttempts:
                opFile.Status = "Failed"
                if opFile.Error is None:
                    opFile.Error = ""
                opFile.Error += " (Max attempts limit reached)"
        return [opFile for opFile in self.operation if opFile.Status == "Waiting"]

    def rssSEStatus(self, se, status, retries=2):
        """check SE :se: for status :status:

        :param str se: SE name
        :param str status: RSS status
        """
        # Allow a transient failure
        for _i in range(retries):
            rssStatus = self.rssClient().getElementStatus(se, "StorageElement", status)
            # gLogger.always( rssStatus )
            if rssStatus["OK"]:
                return S_OK(rssStatus["Value"][se][status] != "Banned")
        return S_ERROR("%s status not found in RSS for SE %s" % (status, se))

    @property
    def shifter(self):
        return self.__shifterList

    @shifter.setter
    def shifter(self, shifterList):
        self.__shifterList = shifterList

    def __call__(self):
        """this one should be implemented in the inherited class

        should return S_OK/S_ERROR
        """
        raise NotImplementedError("Implement me please!")

    def createRMSRecord(self, status, nbObject):
        """
        This method is used to create a record given some parameters for sending it to the ES backend.
        It is used inside DMS/Agent/RequestOperations and this method is designed particularly for file
        type of objects.

        :param status: This can be one of these i.e. Attempted, Failed, or Successful.
        :param nbObject: This is number of objects in question.

        :returns: a dictionary.
        """
        record = {
            "timestamp": int(Time.toEpoch()),
            "host": Network.getFQDN(),
            "objectType": "File",
            "operationType": self.operation.Type,
            "status": status,
            "nbObject": nbObject,
            "parentID": self.operation.OperationID,
        }

        return record



"""
Classes and functions for easier management of the InstalledComponents database
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import re
import datetime
from sqlalchemy import MetaData, Column, Integer, String, DateTime, create_engine, text
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session, relationship
from sqlalchemy.schema import ForeignKey
from sqlalchemy.sql.expression import null

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters

__RCSID__ = "$Id$"

metadata = MetaData()
componentsBase = declarative_base()


class Component(componentsBase):
    """
    This class defines the schema of the Components table in the
    InstalledComponentsDB database
    """

    __tablename__ = "Components"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    componentID = Column("ComponentID", Integer, primary_key=True)
    system = Column("DIRACSystem", String(32), nullable=False)
    module = Column("DIRACModule", String(32), nullable=False)
    cType = Column("Type", String(32), nullable=False)

    def __init__(self, system=null(), module=null(), cType=null()):
        """just defines some instance members"""
        self.system = system
        self.module = module
        self.cType = cType
        self.installationList = []

    def fromDict(self, dictionary):
        """
        Fill the fields of the Component object from a dictionary
        The dictionary may contain the keys: ComponentID, System, Module, Type
        """

        self.componentID = dictionary.get("ComponentID", self.componentID)
        self.system = dictionary.get("DIRACSystem", self.system)
        self.module = dictionary.get("DIRACModule", self.module)
        self.cType = dictionary.get("Type", self.cType)

        return S_OK("Successfully read from dictionary")

    def toDict(self, includeInstallations=False, includeHosts=False):
        """
        Return the object as a dictionary
        If includeInstallations is True, the dictionary returned will also include
        information about the installations in which this Component is included
        If includeHosts is also True, further information about the Hosts where the
        installations are is included
        """

        dictionary = {
            "ComponentID": self.componentID,
            "DIRACSystem": self.system,
            "DIRACModule": self.module,
            "Type": self.cType,
        }

        if includeInstallations:
            installations = []
            for installation in self.installationList:
                relationshipDict = installation.toDict(False, includeHosts)["Value"]
                installations.append(relationshipDict)
            dictionary["Installations"] = installations

        return S_OK(dictionary)


class Host(componentsBase):
    """
    This class defines the schema of the Hosts table in the
    InstalledComponentsDB database
    """

    __tablename__ = "Hosts"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    hostID = Column("HostID", Integer, primary_key=True)
    hostName = Column("HostName", String(32), nullable=False)
    cpu = Column("CPU", String(64), nullable=False)
    installationList = relationship("InstalledComponent", backref="installationHost")

    def __init__(self, host=null(), cpu=null()):
        self.hostName = host
        self.cpu = cpu

    def fromDict(self, dictionary):
        """
        Fill the fields of the Host object from a dictionary
        The dictionary may contain the keys: HostID, HostName, CPU
        """

        self.hostID = dictionary.get("HostID", self.hostID)
        self.hostName = dictionary.get("HostName", self.hostName)
        self.cpu = dictionary.get("CPU", self.cpu)

        return S_OK("Successfully read from dictionary")

    def toDict(self, includeInstallations=False, includeComponents=False):
        """
        Return the object as a dictionary
        If includeInstallations is True, the dictionary returned will also include
        information about the installations in this Host
        If includeComponents is also True, further information about which
        Components where installed is included
        """

        dictionary = {"HostID": self.hostID, "HostName": self.hostName, "CPU": self.cpu}

        if includeInstallations:
            installations = []
            for installation in self.installationList:
                relationshipDict = installation.toDict(includeComponents, False)["Value"]
                installations.append(relationshipDict)
            dictionary["Installations"] = installations

        return S_OK(dictionary)


class InstalledComponent(componentsBase):
    """
    This class defines the schema of the InstalledComponents table in the
    InstalledComponentsDB database
    """

    __tablename__ = "InstalledComponents"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    componentID = Column("ComponentID", Integer, ForeignKey("Components.ComponentID"), primary_key=True)
    hostID = Column("HostID", Integer, ForeignKey("Hosts.HostID"), primary_key=True)
    instance = Column("Instance", String(64), primary_key=True)
    installationTime = Column("InstallationTime", DateTime, primary_key=True)
    unInstallationTime = Column("UnInstallationTime", DateTime)
    installedBy = Column("InstalledBy", String(32))
    unInstalledBy = Column("UnInstalledBy", String(32))
    installationComponent = relationship("Component", backref="installationList")

    def __init__(
        self,
        instance=null(),
        installationTime=null(),
        unInstallationTime=null(),
        installedBy=null(),
        unInstalledBy=null(),
    ):
        self.instance = instance
        self.installationTime = installationTime
        self.unInstallationTime = unInstallationTime
        self.installedBy = installedBy
        self.unInstalledBy = unInstalledBy

    def fromDict(self, dictionary):
        """
        Fill the fields of the InstalledComponent object from a dictionary
        The dictionary may contain the keys: ComponentID, HostID, Instance,
        InstallationTime, UnInstallationTime
        """

        self.componentID = dictionary.get("ComponentID", self.componentID)
        self.hostID = dictionary.get("HostID", self.hostID)
        self.instance = dictionary.get("Instance", self.instance)
        self.installationTime = dictionary.get("InstallationTime", self.installationTime)
        self.unInstallationTime = dictionary.get("UnInstallationTime", self.unInstallationTime)
        self.installedBy = dictionary.get("InstalledBy", self.installedBy)
        self.unInstalledBy = dictionary.get("UnInstalledBy", self.unInstalledBy)

        return S_OK("Successfully read from dictionary")

    def toDict(self, includeComponents=False, includeHosts=False):
        """
        Return the object as a dictionary
        If includeComponents is True, information about which Components where
        installed is included
        If includeHosts is True, information about the Hosts where the
        installations are is included
        """

        dictionary = {
            "Instance": self.instance,
            "InstallationTime": self.installationTime,
            "UnInstallationTime": self.unInstallationTime,
            "InstalledBy": self.installedBy,
            "UnInstalledBy": self.unInstalledBy,
        }

        if includeComponents:
            dictionary["Component"] = self.installationComponent.toDict()["Value"]
        else:
            dictionary["ComponentID"] = self.componentID

        if includeHosts:
            dictionary["Host"] = self.installationHost.toDict()["Value"]
        else:
            dictionary["HostID"] = self.hostID

        return S_OK(dictionary)


class HostLogging(componentsBase):
    """
    This class defines the schema of the HostLogging table in the
    InstalledComponentsDB database
    """

    __tablename__ = "HostLogging"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    hostName = Column("HostName", String(32), nullable=False, primary_key=True)
    # status
    DIRAC = Column("DIRACVersion", String(64))
    Extension = Column("Extension", String(512))
    Load1 = Column("Load1", String(32))  # float
    Load5 = Column("Load5", String(32))  # float
    Load15 = Column("Load15", String(32))  # float
    Memory = Column("Memory", String(32))
    DiskOccupancy = Column("DiskOccupancy", String(512))
    Swap = Column("Swap", String(32))
    CPUClock = Column("CPUClock", String(32))  # float
    CPUModel = Column("CPUModel", String(64))
    CertificateDN = Column("CertificateDN", String(128))
    CertificateIssuer = Column("CertificateIssuer", String(128))
    CertificateValidity = Column("CertificateValidity", String(64))
    Cores = Column("Cores", Integer)
    PhysicalCores = Column("PhysicalCores", Integer)
    OpenFiles = Column("OpenFiles", Integer)
    OpenPipes = Column("OpenPipes", Integer)
    OpenSockets = Column("OpenSockets", Integer)
    Setup = Column("Setup", String(32))
    Uptime = Column("Uptime", String(64))
    Timestamp = Column("Timestamp", DateTime)

    def __init__(self, host=null(), **kwargs):
        self.hostName = host
        fields = dir(self)

        for key, value in kwargs.items():
            if key in fields and not re.match("_.*", key):
                setattr(self, key, value)

    def fromDict(self, dictionary):
        """
        Fill the fields of the HostLogging object from a dictionary
        """
        fields = dir(self)

        if dictionary.get("DIRACVersion"):
            dictionary["DIRAC"] = dictionary.get("DIRACVersion")

        try:
            for key, value in dictionary.items():
                if key in fields and not re.match("_.*", key):
                    setattr(self, key, value)
        except Exception as e:
            return S_ERROR(e)

        return S_OK("Successfully read from dictionary")

    def toDict(self):
        """
        Return the object as a dictionary
        """

        dictionary = {
            "HostName": self.hostName,
            "DIRACVersion": self.DIRAC,
            "Extension": self.Extension,
            "Load1": self.Load1,
            "Load5": self.Load5,
            "Load15": self.Load15,
            "Memory": self.Memory,
            "DiskOccupancy": self.DiskOccupancy,
            "Swap": self.Swap,
            "CPUClock": self.CPUClock,
            "CPUModel": self.CPUModel,
            "CertificateDN": self.CertificateDN,
            "CertificateIssuer": self.CertificateIssuer,
            "CertificateValidity": self.CertificateValidity,
            "Cores": self.Cores,
            "PhysicalCores": self.PhysicalCores,
            "OpenFiles": self.OpenFiles,
            "OpenPipes": self.OpenPipes,
            "OpenSockets": self.OpenSockets,
            "Setup": self.Setup,
            "Uptime": self.Uptime,
            "Timestamp": self.Timestamp,
        }

        return S_OK(dictionary)


class InstalledComponentsDB(object):
    """
    Class used to work with the InstalledComponentsDB database.
    It creates the tables on initialization and allows inserting, querying,
    deleting from/to the tables
    """

    def __init__(self):
        self.__initializeConnection("Framework/InstalledComponentsDB")
        result = self.__initializeDB()
        if not result["OK"]:
            raise Exception("Can't create tables: %s" % result["Message"])

    def __initializeConnection(self, dbPath):

        result = getDBParameters(dbPath)
        if not result["OK"]:
            raise Exception("Cannot get database parameters: %s" % result["Message"])

        dbParameters = result["Value"]
        self.host = dbParameters["Host"]
        self.port = dbParameters["Port"]
        self.user = dbParameters["User"]
        self.password = dbParameters["Password"]
        self.dbName = dbParameters["DBName"]

        self.engine = create_engine(
            "mysql://%s:%s@%s:%s/%s" % (self.user, self.password, self.host, self.port, self.dbName),
            pool_recycle=3600,
            echo_pool=True,
        )
        self.session = scoped_session(sessionmaker(bind=self.engine))
        self.inspector = Inspector.from_engine(self.engine)

    def __initializeDB(self):
        """
        Create the tables
        """

        tablesInDB = self.inspector.get_table_names()

        # Components
        if "Components" not in tablesInDB:
            try:
                Component.__table__.create(self.engine)  # pylint: disable=no-member
            except Exception as e:
                return S_ERROR(e)
        else:
            gLogger.debug("Table 'Components' already exists")

        # Hosts
        if "Hosts" not in tablesInDB:
            try:
                Host.__table__.create(self.engine)  # pylint: disable=no-member
            except Exception as e:
                return S_ERROR(e)
        else:
            gLogger.debug("Table 'Hosts' already exists")

        # InstalledComponents
        if "InstalledComponents" not in tablesInDB:
            try:
                InstalledComponent.__table__.create(self.engine)  # pylint: disable=no-member
            except Exception as e:
                return S_ERROR(e)
        else:
            gLogger.debug("Table 'InstalledComponents' already exists")

        # HostLogging
        if "HostLogging" not in tablesInDB:
            try:
                HostLogging.__table__.create(self.engine)  # pylint: disable=no-member
            except Exception as e:
                return S_ERROR(e)
        else:
            gLogger.debug("Table 'HostLogging' already exists")

        return S_OK("Tables created")

    def __filterFields(self, session, table, matchFields=None):
        """
        Filters instances of a selection by finding matches on the given fields
        session argument is a Session instance used to retrieve the items
        table argument must be one the following three: Component, Host,
        InstalledComponent
        matchFields argument should be a dictionary with the fields to match.
        matchFields accepts fields of the form <Field.bigger> and <Field.smaller>
        to filter using > and < relationships.
        If matchFields is empty, no filtering will be done
        """

        if matchFields is None:
            matchFields = {}

        filtered = session.query(table)

        for key in matchFields:
            actualKey = key

            comparison = "="
            if ".bigger" in key:
                comparison = ">"
                actualKey = key.replace(".bigger", "")
            elif ".smaller" in key:
                comparison = "<"
                actualKey = key.replace(".smaller", "")

            if matchFields[key] is None:
                sql = "`%s` IS NULL" % (actualKey)
            elif isinstance(matchFields[key], list):
                if len(matchFields[key]) > 0 and None not in matchFields[key]:
                    sql = "`%s` IN ( " % (actualKey)
                    for i, element in enumerate(matchFields[key]):
                        toAppend = element
                        if isinstance(toAppend, datetime.datetime):
                            toAppend = toAppend.strftime("%Y-%m-%d %H:%M:%S")
                        if isinstance(toAppend, six.string_types):
                            toAppend = "'%s'" % (toAppend)
                        if i == 0:
                            sql = "%s%s" % (sql, toAppend)
                        else:
                            sql = "%s, %s" % (sql, toAppend)
                    sql = "%s )" % (sql)
                else:
                    continue
            elif isinstance(matchFields[key], six.string_types):
                sql = "`%s` %s '%s'" % (actualKey, comparison, matchFields[key])
            elif isinstance(matchFields[key], datetime.datetime):
                sql = "%s %s '%s'" % (actualKey, comparison, matchFields[key].strftime("%Y-%m-%d %H:%M:%S"))
            else:
                sql = "`%s` %s %s" % (actualKey, comparison, matchFields[key])

            filteredTemp = filtered.filter(text(sql))
            try:
                session.execute(filteredTemp)
                session.commit()
            except Exception as e:
                return S_ERROR("Could not filter the fields: %s" % (e))
            filtered = filteredTemp

        return S_OK(filtered)

    def __filterInstalledComponentsFields(self, session, matchFields=None):
        """
        Filters instances by finding matches on the given fields in the same way
        as the '__filterFields' function
        The main difference with '__filterFields' is that this function is
        targeted towards the InstalledComponents table
        and accepts fields of the form <Component.Field> and <Host.Field>
        ( e.g., 'Component.DIRACSystem' ) to filter installations using attributes
        from their associated Components and Hosts.
        session argument is a Session instance used to retrieve the items
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships.
        matchFields argument should be a dictionary with the fields to match.
        If matchFields is empty, no filtering will be done
        """
        if matchFields is None:
            matchFields = {}
        componentKeys = {}

        for (key, val) in matchFields.items():
            if "Component." in key:
                componentKeys[key.replace("Component.", "")] = val

        hostKeys = {}
        for (key, val) in matchFields.items():
            if "Host." in key:
                hostKeys[key.replace("Host.", "")] = val

        selfKeys = {}
        for (key, val) in matchFields.items():
            if "Component." not in key and "Host." not in key:
                selfKeys[key] = val

        # Get the matching components
        result = self.__filterFields(session, Component, componentKeys)
        if not result["OK"]:
            return result

        componentIDs = []
        for component in result["Value"]:
            componentIDs.append(component.componentID)

        # Get the matching hosts
        result = self.__filterFields(session, Host, hostKeys)
        if not result["OK"]:
            return result

        hostIDs = []
        for host in result["Value"]:
            hostIDs.append(host.hostID)

        # Get the matching InstalledComponents
        result = self.__filterFields(session, InstalledComponent, selfKeys)
        if not result["OK"]:
            return result

        # And use the Component and Host IDs to filter them as well
        installations = (
            result["Value"]
            .filter(InstalledComponent.componentID.in_(componentIDs))
            .filter(InstalledComponent.hostID.in_(hostIDs))
        )

        return S_OK(installations)

    def exists(self, table, matchFields):
        """
        Checks whether an instance matching the given criteria exists
        table argument must be one the following three: Component, Host,
        InstalledComponent
        matchFields argument should be a dictionary with the fields to match.
        If matchFields is empty, no filtering will be done
        matchFields may contain entries of the form 'Component.attribute' or
        'Host.attribute' if table equals InstalledComponent
        """

        session = self.session()

        if table == InstalledComponent:
            result = self.__filterInstalledComponentsFields(session, matchFields)
        else:
            result = self.__filterFields(session, table, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result
        query = result["Value"]

        session.commit()
        session.close()

        if query.count() == 0:
            return S_OK(False)
        else:
            return S_OK(True)

    def addComponent(self, newComponent):
        """
        Add a new component to the database
        newComponent argument should be a dictionary with the Component fields and
        its values

        NOTE: The addition of the items is temporary. To commit the changes to
        the database it is necessary to call commitChanges()
        """

        session = self.session()

        component = Component()
        component.fromDict(newComponent)

        try:
            session.add(component)
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not add Component: %s" % (e))

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("Component successfully added")

    def removeComponents(self, matchFields=None):
        """
        Removes components with matches in the given fields
        matchFields argument should be a dictionary with the fields and values
        to match
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        matchFields argument can be empty to remove all the components

        NOTE: The removal of the items is temporary. To commit the changes to
        the database it is necessary to call commitChanges()
        """
        if matchFields is None:
            matchFields = {}

        session = self.session()

        result = self.__filterFields(session, Component, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        for component in result["Value"]:
            session.delete(component)

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("Components successfully removed")

    def getComponents(self, matchFields={}, includeInstallations=False, includeHosts=False):
        """
        Returns a list with all the components with matches in the given fields
        matchFields argument should be a dictionary with the fields to match or
        empty to get all the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        includeInstallations indicates whether data about the installations in
        which the components takes part is to be retrieved
        includeHosts (only if includeInstallations is set to True) indicates
        whether data about the host in which there are instances of this component
        is to be retrieved
        """

        session = self.session()

        result = self.__filterFields(session, Component, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        components = result["Value"]
        if not components:
            session.rollback()
            session.close()
            return S_ERROR("No matching Components were found")

        dictComponents = []
        for component in components:
            dictComponents.append(component.toDict(includeInstallations, includeHosts)["Value"])

        session.commit()
        session.close()
        return S_OK(dictComponents)

    def getComponentByID(self, cId):
        """
        Returns a component given its id
        """

        result = self.getComponents(matchFields={"ComponentID": cId})
        if not result["OK"]:
            return result

        component = result["Value"]
        if component.count() == 0:
            return S_ERROR("Component with ID %s does not exist" % (cId))

        return S_OK(component[0])

    def componentExists(self, component):
        """
        Checks whether the given component exists in the database or not
        """

        session = self.session()

        try:
            query = (
                session.query(Component)
                .filter(text(Component.system == component.system))
                .filter(text(Component.module == component.module))
                .filter(text(Component.cType == component.cType))
            )
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Couldn't check the existence of the component: %s" % (e))

        session.commit()
        session.close()
        if query.count() == 0:
            return S_OK(False)
        else:
            return S_OK(True)

    def updateComponents(self, matchFields={}, updates={}):
        """
        Updates Components objects on the database
        matchFields argument should be a dictionary with the fields to match
        (instances matching the fields will be updated) or empty to update all
        the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships updates argument
        should be a dictionary with the Component fields and their new
        updated values
        updates argument should be a dictionary with the Installation fields and
        their new updated values
        """

        session = self.session()

        result = self.__filterFields(session, Component, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        components = result["Value"]

        for component in components:
            component.fromDict(updates)

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("Component(s) updated")

    def addHost(self, newHost):
        """
        Add a new host to the database
        host argument should be a dictionary with the Host fields and its values

        NOTE: The addition of the items is temporary. To commit the changes to
        the database it is necessary to call commitChanges()
        """

        session = self.session()

        host = Host()
        host.fromDict(newHost)

        try:
            session.add(host)
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not add Host: %s" % (e))

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("Host successfully added")

    def removeHosts(self, matchFields={}):
        """
        Removes hosts with matches in the given fields
        matchFields argument should be a dictionary with the fields and values
        to match
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        matchFields argument can be empty to remove all the hosts

        NOTE: The removal of the items is temporary. To commit the changes to
        the database it is necessary to call commitChanges()
        """

        session = self.session()

        result = self.__filterFields(session, Host, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        for host in result["Value"]:
            session.delete(host)

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("Hosts successfully removed")

    def getHosts(self, matchFields={}, includeInstallations=False, includeComponents=False):
        """
        Returns a list with all the hosts with matches in the given fields
        matchFields argument should be a dictionary with the fields to match or
        empty to get all the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        includeInstallations indicates whether data about the installations in
        which the host takes part is to be retrieved
        includeComponents (only if includeInstallations is set to True) indicates
        whether data about the components installed into this host is to
        be retrieved
        """

        session = self.session()

        result = self.__filterFields(session, Host, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        hosts = result["Value"]
        if not hosts:
            session.rollback()
            session.close()
            return S_ERROR("No matching Hosts were found")

        dictHosts = []
        for host in hosts:
            dictHosts.append(host.toDict(includeInstallations, includeComponents)["Value"])

        session.commit()
        session.close()

        return S_OK(dictHosts)

    def getHostByID(self, cId):
        """
        Returns a host given its id
        """

        result = self.getHosts(matchFields={"HostID": cId})
        if not result["OK"]:
            return result

        host = result["Value"]
        if host.count() == 0:
            return S_ERROR("Host with ID %s does not exist" % (cId))

        return S_OK(host[0])

    def hostExists(self, host):
        """
        Checks whether the given host exists in the database or not
        """

        session = self.session()

        try:
            query = session.query(Host).filter(text(Host.hostName == host.hostName)).filter(text(Host.cpu == host.cpu))
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not check the existence of the host: %s" % (e))

        session.commit()
        session.close()
        if query.count() == 0:
            return S_OK(False)
        else:
            return S_OK(True)

    def updateHosts(self, matchFields={}, updates={}):
        """
        Updates Hosts objects on the database
        matchFields argument should be a dictionary with the fields to
        match (instances matching the fields will be updated) or empty to update
        all the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships updates argument
        should be a dictionary with the Host fields and their new updated values
        updates argument should be a dictionary with the Installation fields and
        their new updated values
        """

        session = self.session()

        result = self.__filterFields(session, Host, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        hosts = result["Value"]

        for host in hosts:
            host.fromDict(updates)

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("Host(s) updated")

    def addInstalledComponent(self, newInstallation, componentDict, hostDict, forceCreate=False):
        """
        Add a new installation of a component to the database
        installation argument should be a dictionary with the InstalledComponent
        fields and its values
        componentDict argument should be a dictionary with the Component fields
        and its values
        hostDict argument should be a dictionary with the Host fields and
        its values
        If forceCreate is set to True, both the component and the host will be
        created if they do not exist

        NOTE: The addition of the items is temporary. To commit the changes to
        the database it is necessary to call commitChanges()
        """

        session = self.session()

        installation = InstalledComponent()
        installation.fromDict(newInstallation)

        result = self.__filterFields(session, Component, componentDict)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result
        if result["Value"].count() != 1:
            if result["Value"].count() > 1:
                session.rollback()
                session.close()
                return S_ERROR("Too many Components match the criteria")
            if result["Value"].count() < 1:
                if not forceCreate:
                    session.rollback()
                    session.close()
                    return S_ERROR("Given component does not exist")
                else:
                    component = Component()
                    component.fromDict(componentDict)
        else:
            component = result["Value"][0]

        result = self.__filterFields(session, Host, hostDict)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result
        if result["Value"].count() == 1:
            host = result["Value"][0]
        elif result["Value"].count() > 1:
            session.rollback()
            session.close()
            return S_ERROR("Too many Hosts match the criteria")
        elif result["Value"].count() == 0:
            if not forceCreate:
                session.rollback()
                session.close()
                return S_ERROR("Given host does not exist")
            # check if HostName exists with different CPU
            gLogger.verbose("Host not found, looking just for hostname")
            hostNameDict = {"HostName": hostDict["HostName"]}
            result = self.__filterFields(session, Host, hostNameDict)
            if result["Value"].count() == 1:
                gLogger.verbose("HostName found, updating CPU model")
                host = result["Value"][0]
                self.updateHosts(hostNameDict, hostDict)
            elif result["Value"].count() > 1:
                session.rollback()
                session.close()
                return S_ERROR("Too many Hosts match the HostName")
            else:
                host = Host()
                host.fromDict(hostDict)

        if component:
            installation.installationComponent = component
        if host:
            installation.installationHost = host

        try:
            session.add(installation)
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not add installation: %s" % (e))

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("InstalledComponent successfully added")

    def getInstalledComponents(self, matchFields=None, installationsInfo=False):
        """
        Returns a list with all the InstalledComponents with matches in the given
        fields
        matchFields argument should be a dictionary with the fields to match or
        empty to get all the instances and may contain entries of the form
        'Component.attribute' or 'Host.attribute'
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        installationsInfo indicates whether information about the components and
        host taking part in the installation is to be provided
        """
        if matchFields is None:
            matchFields = {}

        session = self.session()

        result = self.__filterInstalledComponentsFields(session, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        installations = result["Value"]
        if not installations:
            session.rollback()
            session.close()
            return S_ERROR("No matching Installations were found")

        dictInstallations = []
        for installation in installations:
            dictInstallations.append(installation.toDict(installationsInfo, installationsInfo)["Value"])

        session.commit()
        session.close()

        return S_OK(dictInstallations)

    def updateInstalledComponents(self, matchFields={}, updates={}):
        """
        Updates installations matching the given criteria
        matchFields argument should be a dictionary with the fields to match or
        empty to get all the instances and may contain entries of the form
        'Component.attribute' or 'Host.attribute'
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        updates argument should be a dictionary with the Installation fields and
        their new updated values
        """

        session = self.session()

        result = self.__filterInstalledComponentsFields(session, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        installations = result["Value"]

        for installation in installations:
            installation.fromDict(updates)

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("InstalledComponent(s) updated")

    def removeInstalledComponents(self, matchFields={}):
        """
        Removes InstalledComponents with matches in the given fields
        matchFields argument should be a dictionary with the fields and values
        to match and may contain entries of the form 'Component.attribute'
        or 'Host.attribute'
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships.
        matchFields argument can be empty to remove all the hosts

        NOTE: The removal of the items is temporary. To commit the changes to
        the database it is necessary to call commitChanges()
        """

        session = self.session()

        result = self.__filterInstalledComponentsFields(session, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        installations = result["Value"]

        for installation in installations:
            session.delete(installation)

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("InstalledComponents successfully removed")

    def addLog(self, newLog):
        """
        Add a new log to the database
        newLog argument should be a dictionary with the log fields and
        its values.
        Valid keys for newLog include fields that are present in a HostLogging object:
        HostName, DIRACVersion, Load1, Load5, ... of which only HostName is mandatory
        """
        session = self.session()

        log = HostLogging()
        log.fromDict(newLog)

        try:
            session.add(log)
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not add log: %s" % (e))

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("Log successfully added")

    def removeLogs(self, matchFields={}):
        """
        Removes logs with matches in the given fields
        matchFields argument should be a dictionary with the fields and values
        to match
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        matchFields argument can be empty to remove all the logs
        """

        session = self.session()

        result = self.__filterFields(session, HostLogging, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        for log in result["Value"]:
            session.delete(log)

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("Logs successfully removed")

    def getLogs(self, matchFields={}):
        """
        Returns a list with all the logs with matches in the given fields
        matchFields argument should be a dictionary with the fields to match or
        empty to get all the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        """

        session = self.session()

        result = self.__filterFields(session, HostLogging, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        logs = result["Value"]
        if not logs:
            session.rollback()
            session.close()
            return S_ERROR("No matching logs were found")

        dictLogs = []
        for log in logs:
            dictLogs.append(log.toDict()["Value"])

        session.commit()
        session.close()
        return S_OK(dictLogs)

    def updateLogs(self, matchFields={}, updates={}):
        """
        Updates logs matching the given criteria
        matchFields argument should be a dictionary with the fields to match or
        empty to get all the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        updates argument should be a dictionary with the logs fields and
        their new updated values
        """

        session = self.session()

        result = self.__filterFields(session, HostLogging, matchFields)
        if not result["OK"]:
            session.rollback()
            session.close()
            return result

        logs = result["Value"]

        for log in logs:
            log.fromDict(updates)

        try:
            session.commit()
        except Exception as e:
            session.rollback()
            session.close()
            return S_ERROR("Could not commit changes: %s" % (e))

        session.close()
        return S_OK("Log(s) updated")



""" The CS! (Configuration Service)

Modified to work with Tornado
Encode data in base64 because of JSON limitations
In client side you must use a specific client

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from base64 import b64encode, b64decode

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ConfigurationSystem.private.ServiceInterfaceTornado import ServiceInterfaceTornado as ServiceInterface
from DIRAC.Core.Tornado.Server.TornadoService import TornadoService

sLog = gLogger.getSubLogger(__name__)


class TornadoConfigurationHandler(TornadoService):
    """
    The CS handler
    """

    ServiceInterface = None
    PilotSynchronizer = None

    @classmethod
    def initializeHandler(cls, serviceInfo):
        """
        Initialize the configuration server
        Behind it starts thread which refresh configuration
        """
        cls.ServiceInterface = ServiceInterface(serviceInfo["URL"])
        return S_OK()

    def export_getVersion(self):
        """
        Returns the version of the configuration
        """
        return S_OK(self.ServiceInterface.getVersion())

    def export_getCompressedData(self):
        """
        Returns the configuration
        """
        sData = self.ServiceInterface.getCompressedConfigurationData()
        return S_OK(b64encode(sData).decode())

    def export_getCompressedDataIfNewer(self, sClientVersion):
        """
        Returns the configuration if a newer configuration exists, if not just returns the version

        :param sClientVersion: Version used by client
        """
        sVersion = self.ServiceInterface.getVersion()
        retDict = {"newestVersion": sVersion}
        if sClientVersion < sVersion:
            retDict["data"] = b64encode(self.ServiceInterface.getCompressedConfigurationData()).decode()
        return S_OK(retDict)

    def export_publishSlaveServer(self, sURL):
        """
        Used by slave server to register as a slave server.

        :param sURL: The url of the slave server.
        """
        self.ServiceInterface.publishSlaveServer(sURL)
        return S_OK()

    def export_commitNewData(self, sData):
        """
        Write the new configuration
        """
        credDict = self.getRemoteCredentials()
        if "DN" not in credDict or "username" not in credDict:
            return S_ERROR("You must be authenticated!")
        sData = b64decode(sData)
        return self.ServiceInterface.updateConfiguration(sData, credDict["username"])

    def export_writeEnabled(self):
        """
        Used to know if we can change the configuration on this server
        """
        return S_OK(self.ServiceInterface.isMaster())

    def export_getCommitHistory(self, limit=100):
        """
        Get the history of modifications in the configuration
        """
        if limit > 100:
            limit = 100
        history = self.ServiceInterface.getCommitHistory()
        if limit:
            history = history[:limit]
        return S_OK(history)

    def export_getVersionContents(self, versionList):
        """
        Get an old version of the configuration, can also be used to get more than one version.

        :param versionList: List of the version we are trying to get
        """
        contentsList = []
        for version in versionList:
            retVal = self.ServiceInterface.getVersionContents(version)
            if retVal["OK"]:
                contentsList.append(retVal["Value"])
            else:
                return S_ERROR("Can't get contents for version %s: %s" % (version, retVal["Message"]))
        return S_OK(contentsList)

    def export_rollbackToVersion(self, version):
        """
        Rollback to an older version of the configuration

        :param version: The version we want to apply
        """
        retVal = self.ServiceInterface.getVersionContents(version)
        if not retVal["OK"]:
            return S_ERROR("Can't get contents for version %s: %s" % (version, retVal["Message"]))
        credDict = self.getRemoteCredentials()
        if "DN" not in credDict or "username" not in credDict:
            return S_ERROR("You must be authenticated!")
        return self.ServiceInterface.updateConfiguration(
            retVal["Value"], credDict["username"], updateVersionOption=True
        )



""" Job accounting type.

    Filled by the JobWrapper (by the jobs) and by the agent "WorloadManagement/StalledJobAgent"
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import DIRAC
from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType

__RCSID__ = "$Id$"


class Job(BaseAccountingType):
    def __init__(self):
        super(Job, self).__init__()
        self.definitionKeyFields = [
            ("User", "VARCHAR(64)"),
            ("UserGroup", "VARCHAR(32)"),
            ("JobGroup", "VARCHAR(64)"),
            ("JobType", "VARCHAR(32)"),
            ("JobClass", "VARCHAR(32)"),
            ("ProcessingType", "VARCHAR(256)"),
            ("Site", "VARCHAR(64)"),
            ("FinalMajorStatus", "VARCHAR(32)"),
            ("FinalMinorStatus", "VARCHAR(256)"),
        ]
        self.definitionAccountingFields = [
            ("CPUTime", "INT UNSIGNED"),  # utime + stime + cutime + cstime
            ("NormCPUTime", "INT UNSIGNED"),  # CPUTime * CPUNormalizationFactor
            ("ExecTime", "INT UNSIGNED"),  # elapsed_time (wall time) * numberOfProcessors
            ("InputDataSize", "BIGINT UNSIGNED"),
            ("OutputDataSize", "BIGINT UNSIGNED"),
            ("InputDataFiles", "INT UNSIGNED"),
            ("OutputDataFiles", "INT UNSIGNED"),
            ("DiskSpace", "BIGINT UNSIGNED"),
            ("InputSandBoxSize", "BIGINT UNSIGNED"),
            ("OutputSandBoxSize", "BIGINT UNSIGNED"),
            ("ProcessedEvents", "INT UNSIGNED"),  # unused (normally not filled)
        ]
        self.bucketsLength = [
            (86400 * 8, 3600),  # <1w+1d = 1h
            (86400 * 35, 3600 * 4),  # <35d = 4h
            (86400 * 30 * 6, 86400),  # <6m = 1d
            (86400 * 365, 86400 * 2),  # <1y = 2d
            (86400 * 600, 604800),  # >1y = 1w
        ]

        self.checkType()
        # Fill the site
        self.setValueByKey("Site", DIRAC.siteName())

    def checkRecord(self):
        result = self.getValue("ExecTime")
        if not result["OK"]:
            return result
        execTime = result["Value"]
        if execTime > 33350400:  # 1 year
            return DIRAC.S_ERROR("OOps. More than 1 year of cpu time smells fishy!")
        return DIRAC.S_OK()
